'''
AC_SRA_Methods.py, Geoffrey Weal, 20/11/2018

This script is designed to include all the methods for performing structural recognition 
task in terms of the Atom Comparison version of the Structural Recognition Algorithm.

'''

from asap3.analysis.localstructure import FullCNA
from collections import Counter
import sys
import multiprocessing as mp

######################################################################

def get_atomic_CNA_profile(input_data):
	"""
	This method will obtain the Atomic CNA profile

	:param input_data: Contains the cluster to perform the CNA on at the rCut value rCut. 
	:type  input_data: (Organisms.GA.Cluster, float)

	:returns: The atomic CNA profile.
	:rtype:   Counter
	"""
	(system,rCut) = input_data
	fullCNA_atoms = FullCNA(system,rCut)
	CNA_profile_on_each_atom = fullCNA_atoms.get_normal_cna()
	for index in range(len(CNA_profile_on_each_atom)):
		CNA_profile_of_an_atom = list(CNA_profile_on_each_atom[index].items())
		CNA_profile_of_an_atom.sort(key=lambda sign:sign[0])
		CNA_profile_on_each_atom[index] = str(CNA_profile_of_an_atom)
	return Counter(CNA_profile_on_each_atom)

def get_tasks(system, rCuts):
	"""
	This is a generator that allows many atomic_CNA_profiles to be obtained at many rCut values in parallel
	
	:param system: This is the cluster to obtain the atomic_CNA_profiles for
	:type  system: Organisms.GA.Cluster
	:param rCuts: This is the rCut value to to obtain the atomic_CNA_profiles at.
	:type  rCuts: float

	:returns: Yields a tuple of (the cluster, rCut)
	:rtype:   (Organisms.GA.Cluster, float)

	"""
	for rCut in rCuts:
		yield (system,rCut)

def get_CNA_profile(input_data):
	'''
	This def will return the CNA profile of a cluster at a range of given values of rCut.

	:param cluster: This is the cluster to obtain the CNA profile of.
	:type  cluster: Either Cluster or ase.Atoms
	:param rCuts: The range of values of rCuts to obtain the CNA profile of.
	:type  rCuts: float

	:returns: This resutrn the name of the cluster, and the atomic_CNA_profiles. 
	:rtype:   (int, Counter)
	'''
	cluster, rCuts = input_data
	system = cluster
	manager = mp.Manager()

	counter = 0
	max_task_execution = 80
	CNA_profile = []
	while counter < len(rCuts):
		ind_start = counter
		ind_end   = ind_start + max_task_execution
		if counter > len(rCuts):
			counter = len(rCuts)
		tasks = get_tasks(system, rCuts[ind_start:ind_end])
		return_list = manager.list()
		for task in tasks:
			p = mp.Process(target=get_atomic_CNA_profile, args=(task,return_list))
			p.run()
		counter += max_task_execution
		CNA_profile += list(return_list)
	return (system.name,CNA_profile)

#############################################################################################

def get_CNA_similarity(cluster_1_CNA,cluster_2_CNA,total_no_of_atoms):
	'''
	Get the similarity for the two clusters at a particular value of rCut.

	:param cluster_1_CNA: the CNA profile of cluster 1 at rCut
	:type  cluster_1_CNA: asap3.analysis.localstructure.FullCNA
	:param cluster_2_CNA: the CNA profile of cluster 2 at rCut
	:type  cluster_2_CNA: asap3.analysis.localstructure.FullCNA
	:param total_no_of_atoms: The total number of atoms in the cluster
	:type  total_no_of_atoms: int
	
	'''
	common_atom_signatures = (Counter(cluster_1_CNA) & Counter(cluster_2_CNA))
	total_no_common_atoms = sum(common_atom_signatures.values())
	similarity = float(total_no_common_atoms)/float(total_no_of_atoms)*100.0
	return similarity

def get_CNA_similarities(input_data):
	"""
	Get the full similarity profile of the two clusters.

	input_data contains two parameters

	:param name_1: Name of the first cluster
	:type  name_1: int
	:param name_2: Name of the second cluster
	:type  name_2: int
	:param cluster_1_CNA_profile: the full CNA profile of cluster 1 for all values of rCut.
	:type  cluster_1_CNA_profile: [asap3.analysis.localstructure.FullCNA ,...]
	:param cluster_2_CNA_profile: the full CNA profile of cluster 2 for all values of rCut.
	:type  cluster_2_CNA_profile: [asap3.analysis.localstructure.FullCNA ,...]

	:returns: returns the name of the two clusters, and the similarity profile.
	:rtype:   (int, int, list of float)

	"""
	name_1, name_2, cluster_1_CNA_profile, cluster_2_CNA_profile = input_data
	CNA_similarities = []
	if not sum(cluster_1_CNA_profile[0].values()) == sum(cluster_2_CNA_profile[0].values()):
		print('Error in def get_CNA_similarities, in AC_SRA_Methods.py')
		print('cluster_1_CNA_profile and cluster_2_CNA_profile are not the same size')
		print('They should be the same size as they should have been examined across the same rCut values.')
		print('len(cluster_1_CNA_profile) = '+str(len(cluster_1_CNA_profile)))
		print('len(cluster_2_CNA_profile) = '+str(len(cluster_2_CNA_profile)))
		print('Check this out.')
		import pdb; pdb.set_trace()
		exit()
	total_no_of_atoms = sum(cluster_1_CNA_profile[0].values())
	for index in range(len(cluster_1_CNA_profile)):
		cluster_1_CNA = cluster_1_CNA_profile[index]; cluster_2_CNA = cluster_2_CNA_profile[index]; 
		similarity = get_CNA_similarity(cluster_1_CNA,cluster_2_CNA,total_no_of_atoms)
		CNA_similarities.append(similarity)
	return name_1, name_2, CNA_similarities



import os
import numpy as np
from copy import deepcopy
from Organisms.GA.Predation_Operators.Predation_Operator import Predation_Operator
from Organisms.GA.SCM_Scripts.CNA_Database import CNA_Database
from Organisms.GA.SCM_Scripts.SCM_initialisation import get_SCM_methods, get_rCut_values, get_rCuts

class Cluster_Block:
	"""
	This is used by the Remove_Cluster_Due_To_Predation_Violation definition to store information in an easy way to help the user understand what is going on in this method 

	:param collection: This is the collection to record. This is either the instance of the Population or the Offspring_Pool
	:type  collection: Population or Offspring_Pool
	:param collection_type: This describes if the cluster recorded is in the population (given as 'pop') or the offspring (given as 'off').
	:type  collection_type: str.
	:param index: The position/index of the cluster in the collection.
	:type  index: int

	"""
	def __init__(self,collection,collection_type,index):
		self.collection_type = collection_type
		self.index = index
		self.name = collection[index].name
		self.fitness = collection[index].fitness
		self.energy = collection[index].energy
	def  __repr__(self):
		return str(self.collection_type)+str(self.name)+'('+str(self.index)+','+str(round(self.energy,6))+' eV,'+str(round(self.fitness,3))+')'

class SCM_Predation_Operator(Predation_Operator):
	"""
	This predation operator uses the similarities from the SCM to determine whether to exclude clusters from the population because they are too similar to each other.	

	:param Predation_Information: This contains all the information needed by the Predation Operator you want to use to run.
	:type  Predation_Information: dict.
	:param fitness_information: This is all the settings needed for the SCM predation operator. This is needed if the fitness operator is the structure + energy fitness operator, where the CNA Database maybe the same database for the predation and fitness operator.
	:type  fitness_information: dict.
	:param population: This is the population that this Operator will be controlling to make sure that no two clusters in the population have the same energy.
	:type  population: Organisms.GA.Population
	:param no_of_cpus: This is the number of cpus to use.
	:type  no_of_cpus: int
	:param print_details: Print details of the predation operator, like verbose
	:type  print_details: bool

	"""
	def __init__(self,Predation_Information,fitness_information,population,no_of_cpus,print_details): #,no_generations):
		super().__init__(Predation_Information,population,print_details)
		#self.does_population_need_after_natural_selection_processing = True
		print('######################################################')
		print('SCM Based Predation Algorithm')
		print('')
		############################################################
		# Check the Predation Operator is the correct one for this Predation Operator.
		if not Predation_Information['Predation Operator'] == 'SCM':
			print('Error in class SCM_Predation_Operator, in SCM_Predation_Operator.py')
			print('The Predation Operator "SCM_Predation_Operator" has been initialised.')
			print("However, Predation_Information['Predation Operator'] is not 'SCM'")
			print("Predation_Information['Predation Operator'] = "+str(Predation_Information['Predation Operator']))
			print('Check this out.')
			import pdb; pdb.set_trace()
			exit()
		############################################################
		# Determine which SCM Scheme to use, either the AC-SRA or the TC-SRA.
		self.SCM_Scheme = Predation_Information['SCM Scheme']
		if not self.SCM_Scheme in ['A-SCM','T-SCM']:
			print('Error in class SCM_Predation_Scheme, in SCM_Predation_Scheme.py')
			print('The "SCM_Scheme" option should be either "A-SCM" or "T-SCM"')
			print('You have entered '+str(self.SCM_Scheme))
			print('Check this.')
			exit('This program will finish without running.')
		get_CNA_profile, get_CNA_similarities = get_SCM_methods(self.SCM_Scheme)
		get_cna_profile_method = get_CNA_profile
		get_cna_similarities_method = get_CNA_similarities
		############################################################
		# Get the geometrically similar cut off percentage.
		try:
			self.cut_off_similarity = Predation_Information['Cut_off']
		except:
			self.cut_off_similarity = 100.0
		############################################################
		# Obtain the values for values of rCut to scan across and get the values of rCut the user wishes to investigate
		self.rCuts = get_rCuts(self,Predation_Information)
		############################################################
		# Debugging switch
		if not 'debug' in Predation_Information:
			self.debug = True
		else:
			self.debug = Predation_Information['debug']
		####################################################################################
		# setup the CNA_Database for recording CNA profiles and similarity profiles.
		self.cna_database = CNA_Database(self.rCuts,self.population,self.cut_off_similarity,get_cna_profile_method,get_cna_similarities_method,no_of_cpus,self.debug)
		####################################################################################
		#if not 'Use Same CNA Database' in fitness_information:
		#	fitness_information['Use Same CNA Database'] = False
		if fitness_information['Fitness Operator'] == 'SCM + Energy' and fitness_information['Use Predation Information'] == True:
			self.use_same_CNA_database = True
		else:
			self.use_same_CNA_database = False

	###########################################################################################################
	########################## Methods Required for the def check_initial_population ##########################
	###########################################################################################################

	def check_initial_population(self,return_report=False):
		"""
		This definition is responsible for making sure that the initialised population obeys the CNA Predation Operator. 

		:param return_report: Will return a dict with all the information about what clusters are similar to what other clusters in the population. 
		:type  return_report: bool.

		returns:
			* **clusters_to_remove** (*list of ints*): a list of the clusters to remove from the population as they violate the Predation Operator. Format is [(index in population, name fo cluster),...]
			* **CNA_report** (*dict.*): a dictionary with information on the clusters being removed and the other clusters in the population which have caused the violation to the SCM Predation Operator. This information is only used to display information so they know why there are violations to the Predation Operator when they occur. For is {removed cluster: [list of clusters that this cluster is similar to in the population.]}
		
		"""
		###############################################################################
		# First, the CNA_Database is built from the clusters that are in the newly inialised population.
		# 1.1 - get the 
		if self.use_same_CNA_database == False:
			self.cna_database.add(self.population,initialise=True)
		###############################################################################
		if return_report:
			clusters_to_remove, CNA_report = self.get_similar_clusters_to_remove(return_report)
		else:
			clusters_to_remove = self.get_similar_clusters_to_remove(return_report)
		###############################################################################
		# SECOND: Sort and remove cluster from the population and the cna_database for being geometrically similar.
		clusters_to_remove.sort(key=lambda x:x[0], reverse=False) # sort index from lowest to highest
		to_removes = []
		for pop_index_to_remove, name_to_remove in clusters_to_remove[::-1]: #remove index from highest to lowest
			self.population.remove(pop_index_to_remove)
			to_removes.append(name_to_remove)
		self.cna_database.remove(to_removes) # delete the cluster entrance in the self.cna_database
		###############################################################################
		# Return the information, depending on the value fo return_report
		if return_report:
			return clusters_to_remove, CNA_report
		else:
			return clusters_to_remove
		###############################################################################

	def get_similar_clusters_to_remove(self,return_report):
		"""
		Will update the CNA_Database and return all the names of clusters in the population that violate the SCM predation operator

		:param return_report: Will return a dict with all the information about what clusters are similar to what other clusters in the population. 
		:type  return_report: bool.

		"""
		# Determine which clusters are too similar to other clusters in the population.
		# This will give a dictionary which is written by {cluster_dir: [list of other clusters that are geometrically similar to this cluster_dir]}
		geometrically_similar_clusters = self.cna_database.get_similar_clusters_in_database()
		###############################################################################
		# convert geometrically_similar_clusters to list and order clusters based on their energies
		geometrically_similar_clusters = list(map(list, geometrically_similar_clusters.items())) # given as [[dir, list of sim clusters, dir fitness]]
		for index in range(len(geometrically_similar_clusters)):
			cluster_name = geometrically_similar_clusters[index][0]
			cluster_fitness = self.population.get_cluster_from_name(cluster_name).fitness
			geometrically_similar_clusters[index].append(cluster_fitness)
		###############################################################################
		# sort geometrically_similar_clusters by the fitness of the cluster
		# order fitnesses,From highest to lowest
		geometrically_similar_clusters.sort(key=lambda x:x[2], reverse=True)
		if (not geometrically_similar_clusters == []) and (not geometrically_similar_clusters[0][2] >= geometrically_similar_clusters[-1][2]):
			print('Error in def get_similar_clusters_to_remove in class CNA_Diversity_Scheme, in Energy_Diversity_Scheme.py')
			print('The fitnesses after the sort are the wrong way around')
			print('geometrically_similar_clusters[0][2] (fitness) = '+str(geometrically_similar_clusters[0][2]))
			print('geometrically_similar_clusters[-1][2] (fitness)= '+str(geometrically_similar_clusters[-1][2]))
			print('The fitnesses need to be order from highest to lowest.')
			exit('Check this out. This program will finish without completing.')
		###############################################################################
		if return_report: # make a record of clusters that were kepted, as well as the other clusters that were removed for being geometrically similar.
			CNA_report = {}
		clusters_to_remove = []
		index1 = 0
		while index1 < len(geometrically_similar_clusters):
		#for index1 in range(len(geometrically_similar_clusters)):
			cluster_name, other_cluster_names_similar_too, fitness_NOTUSED = geometrically_similar_clusters[index1]
			if return_report:
				CNA_report[cluster_name] = deepcopy(other_cluster_names_similar_too)
			for name_to_remove in other_cluster_names_similar_too:
				clusters_to_remove.append([self.population.get_index(name_to_remove),name_to_remove]) 
				# remove every instance of name_to_remove in rest of geometrically_similar_clusters.
				for index2 in range(len(geometrically_similar_clusters)-1,index1,-1):
					if name_to_remove == geometrically_similar_clusters[index2][0]:
						del geometrically_similar_clusters[index2]
					elif name_to_remove in geometrically_similar_clusters[index2][1]:
						geometrically_similar_clusters[index2][1].remove(name_to_remove)
			index1 += 1
		###############################################################################
		if return_report:
			return clusters_to_remove, CNA_report
		else:
			return clusters_to_remove

	########################################################################################################
	########################## Methods Required for the def assess_for_violations ##########################
	########################################################################################################

	def assess_for_violations(self,offspring_pool,force_replace_pop_clusters_with_offspring):
		"""
		This definition is designed to determine which offspring (and the clusters in the population) violate the Predation Operator. 
		It will not remove or change any clusters in the offspring or population, but instead will record which offspring violate the 
		Predation Operator. 

		It will return two tuples with nothing in them, as required by this def. 

		:param offspring_pool: This is the collection of offspring to assess for violations to the Predation Operator.
		:type  offspring_pool: Organisms.GA.Offspring_Pool.Offspring_Pool
		:param force_replace_pop_clusters_with_offspring: This will tell the genetic algorithm whether to swap clusters in the population with offspring if the predation operator indicates they are the same but the predation operator has a better fitness value than the cluster in the population. 
		:type  force_replace_pop_clusters_with_offspring: bool.

		returns:
			* **offspring_to_remove** (*tuple of ints*): A list of the names of the offspring to be removed
			* **force_replacement** (*tuple of (int, int)*): A list of the clusters in the population that should be replaced, and the offspring they should be replaced by.

		"""
		if force_replace_pop_clusters_with_offspring:
			return self.assess_for_violations_force_replacement(offspring_pool)
		else:
			return self.assess_for_violations_no_force_replacement(offspring_pool)

	def assess_for_violations_no_force_replacement(self,offspring_pool):
		"""
		This definition is designed to determine which offspring (and the clusters in the population) violate the Predation Operator. 
		It will not remove or change any clusters in the offspring or population, but instead will record which offspring violate the 
		Predation Operator. 

		It will return two tuples with nothing in them, as required by this def. 

		:param offspring_pool: This is the collection of offspring to assess for violations to the Predation Operator.
		:type  offspring_pool: Organisms.GA.Offspring_Pool.Offspring_Pool

		returns:
			* **offspring_to_remove** (*tuple of ints*): A list of the names of the offspring to be removed
			* **force_replacement** (*tuple of (int, int)*): A list of the clusters in the population that should be replaced, and the offspring they should be replaced by.

		"""
		if self.print_details:
			print('**********************************************************')
			print('**********************************************************')
			print('Offspring removed from the Offspring Pool due to violating the Comprehensive Energy Diversity Scheme.')
			print('NOTE: energy difference minimum = '+str(self.minimum_energy_diff)+' eV.')
			print('**********************************************************')
			print('**********************************************************')
		########################################################################################################
		if self.print_details:
			print('Start assessing for SCM Diversity')
		offspring_to_remove = []
		# Check offspring against themselves
		structures_to_consider = [Cluster_Block(offspring_pool,'Offspring',index) for index in range(len(offspring_pool))]
		structures_to_consider.sort(key=lambda cluster: cluster.fitness, reverse=False)
		index1 = 0
		while index1 < len(structures_to_consider):
			offspring1 = structures_to_consider[index1]
			for index2 in range(len(structures_to_consider)-1,index1,-1):
				offspring2 = structures_to_consider[index2]
				if self.cna_database.get_max_similarity(offspring1.name,offspring2.name) >= self.cut_off_similarity:
					offspring_to_remove.append((offspring2.name,offspring2.index))
					del structures_to_consider[index2]
			index1 += 1
		# check offspring against clusters in the populatino
		for index in range(len(structures_to_consider)):
			offspring = structures_to_consider[index]
			for cluster_in_population in self.population:
				if self.cna_database.get_max_similarity(offspring.name,cluster_in_population.name) >= self.cut_off_similarity:
					offspring_to_remove.append((offspring.name,offspring.index))
					break
		# Sort by index and continue on
		offspring_to_remove.sort(key=lambda x:x[1],reverse=True)
		force_replacement = []
		return offspring_to_remove, force_replacement

	def assess_for_violations_force_replacement(self,offspring_pool):
		"""
		This definition is designed to determine which offspring (and the clusters in the population) violate the Predation Operator. 
		It will not remove or change any clusters in the offspring or population, but instead will record which offspring violate the 
		Predation Operator. 

		It will return two tuples with nothing in them, as required by this def. 

		:param offspring_pool: This is the collection of offspring to assess for violations to the Predation Operator.
		:type  offspring_pool: Organisms.GA.Offspring_Pool.Offspring_Pool

		returns:
			* **offspring_to_remove** (*tuple of ints*): A list of the names of the offspring to be removed
			* **force_replacement** (*tuple of (int, int)*): A list of the clusters in the population that should be replaced, and the offspring they should be replaced by.

		"""
		###############################################################################
		# If you are using a different fitness Operator and different CNA parameters, you will
		# need to add the offspring in the offspring_pool to the self.cna_database
		#if self.use_same_CNA_database == False:
		#	self.cna_database.add(offspring_pool,initialise=False)
		########################################################################################################
		if self.print_details:
			print('Assessing which structures to remove to comply with SCM Predation Operator')
		structures_to_consider = [Cluster_Block(self.population,'Population',index) for index in range(len(self.population))] 
		structures_to_consider += [Cluster_Block(offspring_pool,'Offspring',index) for index in range(len(offspring_pool))]
		# order fitnesses,From highest to lowest
		structures_to_consider.sort(key=lambda cluster: cluster.fitness, reverse=True)
		if (not structures_to_consider == []) and (not structures_to_consider[0].fitness >= structures_to_consider[-1].fitness):
			print('Error in def assess_for_violations in class CNA_Diversity_Scheme, in Energy_Diversity_Scheme.py')
			print('The fitnesses after the sort are the wrong way around')
			print('geometrically_similar_clusters[0].fitness = '+str(structures_to_consider[0].fitness))
			print('geometrically_similar_clusters[-1].fitness= '+str(structures_to_consider[-1].fitness))
			print('The fitnesses need to be order from highest to lowest.')
			exit('Check this out. This program will finish without completing.')
		########################################################################################################
		########################################################################################################
		# Follows same rough algorithm guide as the comprehensive energy Predation Operator
		if self.print_details:
			print('Start assessing')
		index_higher = 0
		removal_O_P = []; removal_O_O = []
		swap_P_O = []; 
		remove_O_similar_to_too_many_P = []
		# scan though all the clusters in offspring and population in order of fitness
		while index_higher < len(structures_to_consider):
			################################################################################################################
			incountered_population_counter = 0
			cluster_higher = structures_to_consider[index_higher]
			one_instance_of_swap_P_O = []
			removal_O_P_temp = []
			remove_O_similar_to_too_many_P_temp = []
			removal_O_O_temp = []
			# scan though all the other clusters which have higher fitnesses
			indices_to_delete = []
			for index_lower in range(index_higher+1, len(structures_to_consider)):
				cluster_lower = structures_to_consider[index_lower]
				############################################################################################################
				#if self.CNA_database.CNA_Analysis(cluster_higher.dir,cluster_lower.dir) == 'geometric':
				if self.cna_database.get_max_similarity(cluster_higher.name,cluster_lower.name) >= self.cut_off_similarity:
					# The two clusters being looked at at the moment are geometrically the same. 
					if cluster_higher.collection_type == 'Population':
						# if two clusters in the population are geometrically similar, we have an issue.
						if cluster_lower.collection_type == 'Population':
							print('Error in def assess_for_violations in class SCM_Predation_Operator, in SCM_Predation_Operator.py')
							print('Two of the clusters in the population are identical by the max  similarity cutt off setting.')
							print('cluster_higher: '+str(cluster_higher))
							print('cluster_lower: '+str(cluster_lower))
							import pdb; pdb.set_trace()
							exit('This program will exit without completing.')
						# If a cluster in the population (and has a higher fitness) is geometrically similar to an offspring,
						# then it is better to keep the offspring and remove the offspring
						elif cluster_lower.collection_type == 'Offspring':
							removal_O_P_temp.append([cluster_lower,cluster_higher])
					elif cluster_higher.collection_type == 'Offspring':
						# If an offspring is geometrically similar (and has a higher fitness) to a cluster in population,
						# then it is better to replace this cluster in the population with this offspring.
						if cluster_lower.collection_type == 'Population':
							# Take the lowest fitness cluster in the population to swap with this offspring.
							if one_instance_of_swap_P_O == []:
								one_instance_of_swap_P_O = [[cluster_lower,cluster_higher]]
							else:
								cluster_lower_old, cluster_higher_old = one_instance_of_swap_P_O[0]
								if not cluster_higher_old.name == cluster_higher.name:
									exit('Somethings up')
								remove_O_similar_to_too_many_P_temp.append([cluster_higher,[cluster_lower_old,cluster_lower]])
								one_instance_of_swap_P_O = []; indices_to_delete = []
								break
						# If an offspring is geometrically similar (and has a higher fitness) to an offspring,
						# remove the higher fitness offspring
						elif cluster_lower.collection_type == 'Offspring':
							removal_O_O_temp.append([cluster_lower,cluster_higher])
					indices_to_delete.append(index_lower)
			# Add the relavent clusters to the various lists. 
			if not remove_O_similar_to_too_many_P_temp == []:
				remove_O_similar_to_too_many_P += remove_O_similar_to_too_many_P_temp
				del structures_to_consider[index_higher]
			else:
				removal_O_P += removal_O_P_temp
				removal_O_O += removal_O_O_temp
				swap_P_O += one_instance_of_swap_P_O
				for index_lower in sorted(indices_to_delete,reverse=True): # remove from highest to lowest index
					del structures_to_consider[index_lower]
				index_higher += 1
		####################################################################################################################
		# write the list offspring_to_remove that contains all the clusters to be removed from the offspring.
		print('Removing offspring that violate the SCM Predation Operator')
		offspring_to_remove = []
		for off_to_remove, pop_to_keep in sorted(removal_O_P+removal_O_O+remove_O_similar_to_too_many_P,key=lambda x:x[0].index,reverse=True): # remove from highest to lowest index
			if offspring_pool[off_to_remove.index].name in [offspring.name for cluster,offspring in swap_P_O]:
				print('Error in def assess_for_violations, in class SCM_Predation_Operator, in SCM_Predation_Operator.py')
				print('Somehow, this algorithm has identified a structure to be in the offspring_pool and the population')
				print('The structure can only be in either the offspring pool or population.')
				print('Offspring: '+str(offspring_pool[off_to_remove.index].name))
				print('Check this out')
				import pdb; pdb.set_trace()
				exit('This program will end without completing.')
			offspring_to_remove.append((offspring_pool[off_to_remove.index].name,off_to_remove.index))
		if not offspring_to_remove == []:
			print(offspring_to_remove)
		####################################################################################################################
		if self.print_details:
			self.assess_for_violations_message(swap_P_O,remove_O_similar_to_too_many_P,removal_O_P,removal_O_O)
		####################################################################################################################
		# Deal with the offspring that you would like to swap with clusters in the population.
		force_replacement = []
		for pop_to_swap_out, off_to_swap_in in swap_P_O:
			if not off_to_swap_in.name in offspring_pool.get_cluster_names():
				print('temp error in CNA_version_3.py')
				import pdb; pdb.set_trace()
				exit()
			force_replacement.append((pop_to_swap_out.name,off_to_swap_in.name))
		force_replacement.sort(key=lambda x:x[1],reverse=True) # force replace from highest to lowest index
		return offspring_to_remove, force_replacement

	def assess_for_violations_message(self,swap_P_O,remove_O_similar_to_too_many_P,removal_O_P,removal_O_O):
		"""
		This method will tell the information that offspring are removed or replaced and why. 

		:param swap_P_O: The clusters in the population to swap out, and the offspring to swap in its place. These will be replaced because the offspring has a higher fitness than the cluster in the population, even though they are similar. 
		:type  swap_P_O: tuple of (Organisms.GA.SCM_Predation_Operator.Cluster_Block, Organisms.GA.SCM_Predation_Operator.Cluster_Block)
		:param remove_O_similar_to_too_many_P: Remove offspring because there are too many similar clusters to it in the population already. 
		:type  remove_O_similar_to_too_many_P: tuple of (Organisms.GA.SCM_Predation_Operator.Cluster_Block, Organisms.GA.SCM_Predation_Operator.Cluster_Block)
		:param removal_O_P: Remove offspring because they are similar to a cluster in the population. These offspring are less fit than their similar counterparts in the population. 
		:type  removal_O_P: tuple of (Organisms.GA.SCM_Predation_Operator.Cluster_Block, Organisms.GA.SCM_Predation_Operator.Cluster_Block)
		:param removal_O_O: Remove offpsring because they are similar to other clusters in the Offspring_Pool. This offspring is less fit than their similar counterparts in the offspring_pool. 
		:type  removal_O_O: tuple of (Organisms.GA.SCM_Predation_Operator.Cluster_Block, Organisms.GA.SCM_Predation_Operator.Cluster_Block)

		"""
		########################################################################################################
		# Mention offspring that have been removed and due to having a similar energy to which cluster in the population
		# Error with this one
		print('1. Clusters in the Population that have been replaced by Offspring, due to the cluster in the offspring being geometrically similar and having a lower fitness than a offspring.')
		if len(swap_P_O) > 0:
			print('1. Clusters in the Population that have been replaced by Offspring, due to the cluster in the offspring being geometrically similar and having a lower fitness than a offspring.')
			for pop_to_swap_out, off_to_swap_in in sorted(swap_P_O,key=lambda x:x[1].name,reverse=False):
				cp_removed_name = pop_to_swap_out.name; cp_removed_fitness = pop_to_swap_out.fitness; cp_removed_energy = pop_to_swap_out.energy
				off_swapped_in_name = off_to_swap_in.name; off_swapped_in_fitness = off_to_swap_in.fitness; off_swapped_in_energy = off_to_swap_in.energy
				print(' --> Population Cluster removed '+str(cp_removed_name)+' (Energy: '+str(round(cp_removed_energy,6))+' eV; fitness: '+str(round(cp_removed_fitness,6))+') {Replaced with Offspring '+str(off_swapped_in_name)+' (Energy: '+str(round(off_swapped_in_energy,6))+' eV; fitness: '+str(round(off_swapped_in_fitness,6))+')}.')
		else:
			print('1. No clusters in the population have been replaced by offspring in the offsprings list')
		print('*****************************')
		########################################################################################################
		# Mention offspring that have been removed and due to having a similar energy to which cluster in the population
		print('2. These offspring have a better fitness and are geometrically similar to two or more clusters in the population.')
		if len(remove_O_similar_to_too_many_P) > 0:
			print('2. These offspring have a better fitness and are geometrically similar to two or more clusters in the population.')
			print('   Therefore it is best to remove the offspring to avoid having geometrically similar clusters in the population.')
			print('   Below are the offspring that will be removed and the clusters in the population they are geometrically similar to.')
			for off_to_swap_out, geo_sim_pops in sorted(remove_O_similar_to_too_many_P,key=lambda x:x[0].name,reverse=False):
				off_removed_name = off_to_swap_out.name; off_removed_fitness = off_to_swap_out.fitness; off_removed_energy = off_to_swap_out.energy
				geo_sim_pop1, geo_sim_pop2 = geo_sim_pops
				geo_sim_pop1_name = geo_sim_pop1.name; geo_sim_pop1_fitness = geo_sim_pop1.fitness; geo_sim_pop1_energy = geo_sim_pop1.energy
				geo_sim_pop2_name = geo_sim_pop2.name; geo_sim_pop2_fitness = geo_sim_pop2.fitness; geo_sim_pop2_energy = geo_sim_pop2.energy
				print(' --> Offspring removed '+str(off_removed_name)+' (Energy: '+str(round(off_removed_energy,6))+' eV; fitness: '+str(round(off_removed_fitness,6))+') {Geometrically similar to (at least) clusters in population: '+str(geo_sim_pop1_name)+' (Energy: '+str(round(geo_sim_pop1_energy,6))+' eV; fitness: '+str(round(geo_sim_pop1_fitness,6))+')} and '+str(geo_sim_pop2_name)+' (Energy: '+str(round(geo_sim_pop2_energy,6))+' eV; fitness: '+str(round(geo_sim_pop2_fitness,6))+')}.')
		else:
			print('2. No offspring need to be removed from the offsprings list')
		print('*****************************')
		########################################################################################################
		# Mention offspring that have been removed and due to having a similar energy to which cluster in the population
		print('3. The following offspring were removed from the offspring pool (and because they are geometrically similar and have a lower fitness to which cluster in the population):')
		if len(removal_O_P) > 0:
			print('3. The following offspring were removed from the offspring pool (and because they are geometrically similar and have a lower fitness to which cluster in the population):')
			for offspring_removed, cp_due_to in removal_O_P:
				offspring_removed_name = offspring_removed.name; offspring_removed_fitness = offspring_removed.fitness; offspring_removed_energy = offspring_removed.energy
				cp_due_to_name = cp_due_to.name; cp_due_to_fitness = cp_due_to.fitness; cp_due_to_energy = cp_due_to.energy
				print(' --> Offspring '+str(offspring_removed_name)+' (Energy: '+str(offspring_removed_energy)+' eV; fitness: '+str(round(offspring_removed_fitness,6))+') {Similar to Population Cluster '+str(cp_due_to_name)+' (Energy: '+str(round(cp_due_to_energy,6))+' eV; fitness: '+str(round(cp_due_to_fitness,6))+')}.')
		else:
			print('3. No offspring need to be removed from the offsprings list')
		print('*****************************')
		########################################################################################################
		# Mention offspring that have been removed and due to having a similar energy to which other offspring in the Offspring_Pool
		print('4. Offspring that have have been removed for being geometrically similar and have a lower fitness to the rest of the other Offspring')
		if len(removal_O_O) > 0:
			print('4. Offspring that have have been removed for being geometrically similar and have a lower fitness to the rest of the other Offspring')
			for offspring_removed, due_to_offspring in removal_O_O:
				offspring_removed_name = offspring_removed.name; offspring_removed_fitness = offspring_removed.fitness; offspring_removed_energy = offspring_removed.energy
				due_to_offspring_name = due_to_offspring.name; due_to_offspring_fitness = due_to_offspring.fitness; due_to_offspring_energy = due_to_offspring.energy
				print(' --> Offspring '+str(offspring_removed_name)+' (Energy: '+str(offspring_removed_energy)+'eV; fitness: '+str(round(offspring_removed_fitness,6))+') {Similar to Offspring '+str(due_to_offspring_name)+' (Energy: '+str(due_to_offspring_energy)+' eV; fitness: '+str(round(due_to_offspring_fitness,6))+')}.')
		else:
			print('4. No offspring have been removed from the rest of the offsprings list for being too similar')
		print('**********************************************************')
		print('**********************************************************')

	def add_to_database(self, collection):
		'''
		Add clusters similarities to the CNA database to be stored for future generations. 

		:param collection: update the fitnesses of clusters in the collection.
		:type  collection: Organisms.GA.Collection.Collection
		'''
		# If you are using a different fitness Operator and different CNA parameters, you will
		# need to add the offspring in the offspring_pool to the self.cna_database
		if self.use_same_CNA_database == False:
			self.cna_database.add(collection,initialise=False)

	def remove_from_database(self, cluster_names_to_remove):
		'''
		Clusters to remove from the CNA database

		:param cluster_names_to_remove: A list of the names of all the clusters to remove from the CNA database
		:type  cluster_names_to_remove: list of ints
		'''
		if self.use_same_CNA_database == False:
			#for cluster_name in cluster_names_to_remove:
			self.cna_database.remove(cluster_names_to_remove) # delete the cluster entrance in the self.cna_database

	def reset(self):
		'''
		Reset the CNA database with no inputs
		'''
		if self.use_same_CNA_database == False:
			self.cna_database.reset()



'''
Energy_Diversity_Scheme.py, Geoffrey Weal, 28/10/2018

This is one of the Diversity schemes that can be used by the genetic algorithm program. This is the Energy Diversity Scheme.

This scheme works by preventing the population from having clusters with the same energy existing in the population at any one time.

'''

############################################################################################################################################################
###### These Methods are used by other Diversity Scheme. For this reason, these methods have been written to be accessible to those Diversity Schemes ######
############################################################################################################################################################

import os, sys, inspect

###########################################################################################################
########################## Methods Required for the def check_initial_population ##########################
###########################################################################################################

def check_initial_population(self,return_report=False):
	"""
	This definition is responsible for making sure that the initialised population obeys the Energy Diversity Scheme. 

	Here, the clusters in the population are checked to see if they have the same energy. This is needed for after the population has been initally populated with randomly generated clusters and clusters the user has specified.
	
	:param return_report: 
	:type  return_report: bool.

	returns:
		* **clusters_to_remove** (*[[int,str],...]*): Contains a list of [the index of cluster in pop, the dir of the cluster] that have been removed from the population as they violate the Energy Diversity Scheme.
		* **same_energies_report** (*{int: [int,...]}*): Contain a more detailed report of what this method has done by returning a dictorary in the format of {dir of kept cluster: [list of dirs of clusters that have been removed because they have the same energy as the cluster that was kept.]}

	"""
	########################################################################################################
	# get some lists ready to help us.
	cluster_indices_in_population = list(range(len(self.population)))
	clusters_to_remove = []; 
	if return_report:
		same_energies_report = {}
	########################################################################################################
	# Find the energetically repeated clusters in the population as defined by the energy diversity scheme
	#
	# NOTE: Here, we MUST use the list "cluster_indices_in_population". This is because the philosophy of the GA program
	# is to keep the clusters added and removed in-place in the population. For more information abpout what in-place means,
	# see `In-place algorithm <https://en.wikipedia.org/wiki/In-place_algorithm>`_
	#	* Note that no clusters are removed that this point in the algorithm. This is how we can use the list 
	#	  "cluster_indices_in_population" to deals with clusters in the population inplace.
	index1 = 0
	while index1 < len(cluster_indices_in_population):
		index_in_pop1 = cluster_indices_in_population[index1]; cluster1 = self.population[index_in_pop1]
		# scan though the rest of the clusters in the population for any of those clusters with the same energy as cluster1
		for index2 in range(len(cluster_indices_in_population)-1,index1,-1):
			index_in_pop2 = cluster_indices_in_population[index2]; cluster2 = self.population[index_in_pop2]
			# if the energy of cluster2 is the same as cluster1, remove cluster2 from the population.
			if round(cluster2.energy,self.round_energy) == round(cluster1.energy,self.round_energy):
				# If you are giving a report of clusters kept and removed, do the following to place information into "same_energies_report".
				if return_report:
					if not cluster1.dir in same_energies_report:
						same_energies_report[cluster1.dir] = [cluster2.dir]
					else:
						same_energies_report[cluster1.dir].insert(0,cluster2.dir)
				# This cluster will eventually be removed from the population. We will for the moment removed it in place by 
				# removing this clusters index in the "cluster_indices_in_population" list.
				clusters_to_remove.append([index_in_pop2,cluster2.dir])
				del cluster_indices_in_population[index2] # i.e. remove index_in_pop2 from the cluster_indices_in_population list.
		index1 += 1
	########################################################################################################
	# Sort the indices of the clusters_to_remove from lowest to highest index so we can at some point 
	# remove the clusters with these indices from the self.population.
	clusters_to_remove.sort(key=lambda x:x[0],reverse=False)
	########################################################################################################
	# Check to see if there are no two clusters in the clusters_to_remove that are the same cluster (same dir)
	check_for_duplicate_dirs = [cluster_dir_to_remove for _,cluster_dir_to_remove in clusters_to_remove]
	if not len(check_for_duplicate_dirs) == len(set(check_for_duplicate_dirs)):
		print('Error in def check_Population in Class Energy_Diversity_Scheme, in Energy_Diversity_Scheme.py')
		print('Two of the clusters dir in the removed cluster list are the same')
		print('This indicates that this cluster has been deleted twice?')
		print('List of cluster dirs in the population: '+str(check_for_duplicate_dirs))
		print('Check this out.')
		import pdb; pdb.set_trace()
		exit()
	########################################################################################################
	# Return the results, depending on if the user wants to know which clusters were removed as they were energetic similar to which cluster in same_energies_report
	if return_report:
		return clusters_to_remove, same_energies_report
	else:
		return clusters_to_remove
	########################################################################################################

########################################################################################################
########################## Methods Required for the def assess_for_violations ##########################
########################################################################################################

def assess_for_violations(self,offspring_pool):
	"""
	This will assess which offspring to remove before the natural selection process.

	The offspring are assessed against clusters in the population. Offspring are removed from the offsprings if:
	* They have an energy equal to another offspring
	* They have an energy equal to another cluster in the population.

	:param offsprings: This is the offspring that you want to add eventually using the natural selection process to the population.
	:type  offsprings: Offspring_Pool

	"""
	########################################################################################################
	#
	details_of_offsprings_that_will_be_removed = []
	########################################################################################################
	# First, we want to remove any offspring with duplicate energies with other offspring.
	# For example, if 4 offspring have the same energy, the first offspring that was made is kept, and the other three are removed.
	# For this reason, the second while is required to go though the whole of the rest of the offspring_pool to check for more than one duplicate offspring.
	index1 = 0
	while index1 < len(offsprings):
		index2 = index1 + 1
		while index2 < len(offsprings):
			if round(offsprings[index1].energy,self.round_energy) == round(offsprings[index2].energy,self.round_energy):
				offspring_similar_to = ['off_'+str(offsprings[index1].dir),offsprings[index1].energy]
				offspring_to_remove = offsprings.pop(index2)
				details_of_offsprings_that_will_be_removed.append([offspring_to_remove,offspring_similar_to])
			else:
				index2 += 1
		index1 += 1
	########################################################################################################
	# Second, we will now compare the offspring with clusters in the poplation. If a offspring has the same energy as a cluster
	# the population, the offspring will be removed from the offspring pool.
	# The for loop only needs to find one time where offsprings[index1].energy == individual.energy, since there SHOULD be only one cluster
	# in the population with that energy. 
	index1 = 0
	while index1 < len(offsprings):
		for individual in self.population:
			if round(offsprings[index1].energy,self.round_energy) == round(individual.energy,self.round_energy):
				offspring_similar_to = ['pop_'+str(individual.dir),individual.energy]
				offspring_to_remove = offsprings.pop(index1)
				# since we have removed the offspring at index1, we so not want to increment index1, as there will be a new offspring now in index1 due to the pop in the previousl line.
				details_of_offsprings_that_will_be_removed.append([offspring_to_remove,offspring_similar_to])
				break
		else: # This offspring energy was not found in the population, we now increase index1 to move onto the next offspring.
			index1 += 1
	########################################################################################################
	print('*****************************')
	print('*****************************')
	if len(details_of_offsprings_that_will_be_removed) > 0:
		print('Removing the following clusters from the offsprings list')
		for offspring_to_remove, offspring_similar_to in details_of_offsprings_that_will_be_removed:
			print('Cluster '+str(offspring_to_remove.dir)+' ('+str(offspring_to_remove.energy)+' eV) {Similar to '+str(offspring_similar_to[0])+' (Energy = '+str(offspring_similar_to[1])+')}')
	else:
		print('No clusters need to be removed from the offsprings list')
	return [], []

################################################################################################################################################################################################################
############################################################### DEBUGGING METHODS ##############################################################################################################################
################################################################################################################################################################################################################

def Check_for_Issue_with_Scheme_with_collection(self,collection):
	"""
	This method allows us to check that the collection does not violate the simple diversity scheme after it has been perform on the collection.

	The collection is either the Population of Offspring_Pool.

	:param collection: This is the collection to record. This is either the instance of the Population or the Offspring_Pool
	:type  collection: Population or Offspring_Pool
		
	"""
	energies = [cluster.energy for cluster in collection]
	if len(energies) != len(set(energies)):
		print('Error with Energy Diversity Scheme.')
		print('There are duplicates in the energy, which should not be happening in the Energy Diversity Scheme.')
		print('Here is a printout of the clusters in '+str(collection.getname()))
		for cluster in collection:
			print('Cluster '+str(cluster)+'; Energy = '+str(cluster.energy)+' eV.')
		print('This likely means there is an error in the Energy Diversity Scheme code. Check this.')
		import pdb; pdb.set_trace()
		exit()

################################################################################################################################################################################################################
################################################################################################################################################################################################################
################################################################################################################################################################################################################



import sys
from abc import ABC, abstractmethod

class Predation_Operator(ABC):
	"""
	This is an abstract class to act as the skeleton of the Predation_Operator class.

	:param Predation_Information: This contains all the information needed by the Predation Operator you want to use to run.
	:type  Predation_Information: dict.
	:param population: This is the population that this Operator will be controlling to make sure that no two clusters in the population have the same energy.
	:type  population: Organisms.GA.Population
	:param print_details: Print details of the predation operator, like verbose
	:type  print_details: bool

	"""
	def __init__(self,predation_information, population, print_details):
		self.predation_information = predation_information
		self.Predation_Switch = predation_information['Predation Operator']
		self.population = population
		self.print_details = print_details
		self.check()

	def check(self):
		"""
		This method will check to make sure that the predation operator is available.
		"""
		if not self.Predation_Switch in ['Off','Energy','IDCM','SCM']:
			error_string =  'Error in Predation_Operator, in __init__ definition.\n'
			error_string += 'Predation_Operator must be either:'
			error_string += '\t"Off" - No Predation Operator is used.\n'
			error_string += '\t"Energy" - Based on no two clusters can have the same energy in the population.\n'
			error_string += '\t"IDCM" - Based on no two clusters can be structurally identical in the population based on the comparison of those two structuress EDM\'. See Manual for more information.\n'
			error_string += '\t"SCM" - No two structures can be Geometrically Similar based on the Structral Comarison Method. See Manual for more information.\n'
			error_string += 'Check this.\n'
			error_string += 'Predation_Operator = ' + str(self.Predation_Switch)+'\n'
			#import pdb, traceback, sys
			#extype, value, tb = sys.exc_info()
			#traceback.print_exc()
			#pdb.post_mortem(tb)
			print(error_string,file=sys.stderr)
			import pdb; pdb.set_trace()
			exit()

	def __repr__(self):
		return str(self.__dict__)
		
	@abstractmethod
	def check_initial_population(self,return_report=False):
		"""
		This definition is responsible for making sure that the initialised population obeys the Predation Operator of interest.

		:param return_report: Will return a dict with all the information about what clusters are similar to what other clusters in the population. 
		:type  return_report: bool.

		returns:
			* **clusters_to_remove** (*list of ints*): a list of the clusters to remove from the population as they violate the Predation Operator. Format is [(index in population, name fo cluster),...]
			* **CNA_report** (*dict.*): a dictionary with information on the clusters being removed and the other clusters in the population which have caused the violation to the SCM Predation Operator. This information is only used to display information so they know why there are violations to the Predation Operator when they occur. For is {removed cluster: [list of clusters that this cluster is similar to in the population.]}
		
		"""
		pass

	@abstractmethod
	def assess_for_violations(self,offspring_pool,force_replace_pop_clusters_with_offspring):
		"""
		This definition is designed to determine which offspring (and the clusters in the population) violate the Predation Operator. 
		It will not remove or change any clusters in the offspring or population, but instead will record which offspring violate the 
		Predation Operator. 

		It will also recommend which clusters in the population should be removed and be replaced by which offspring prior to the 
		natural selection process. Here, the cluster in the population and the offspring will be in violation of each other, however it 
		may be advantageous to keep the offspring rather than the cluster in the population as the offspring is fitter than the cluster
		in the offspring

		:param offspring_pool: This is the collection of offspring to assess for violations to the Predation Operator.
		:type  offspring_pool: Organisms.GA.Offspring_Pool.Offspring_Pool
		:param force_replace_pop_clusters_with_offspring: This will tell the genetic algorithm whether to swap clusters in the population with offspring if the predation operator indicates they are the same but the predation operator has a better fitness value than the cluster in the population. 
		:type  force_replace_pop_clusters_with_offspring: bool.

		returns:
			* **offspring_to_remove** (*tuple of ints*): A list of the names of the offspring to be removed
			* **force_replacement** (*tuple of (int, int)*): A list of the clusters in the population that should be replaced, and the offspring they should be replaced by.

		"""
		pass

	@abstractmethod
	def add_to_database(self, collection):
		'''
		Add clusters similarities to the CNA database to be stored for future generations. 

		:param collection: update the fitnesses of clusters in the collection.
		:type  collection: Organisms.GA.Collection.Collection
		'''
		pass

	@abstractmethod
	def remove_from_database(self, cluster_names_to_remove):
		'''
		Clusters to remove from the CNA database

		:param cluster_names_to_remove: A list of the names of all the clusters to remove from the CNA database
		:type  cluster_names_to_remove: list of ints
		'''
		pass

	@abstractmethod
	def reset(self):
		'''
		Reset the CNA database with no inputs
		'''
		pass

	def remove_offspring_and_replace_with_population_that_violate_predation_operator(self, population, offspring_pool, offspring_to_remove, population_to_be_replaced_by_offspring):
		"""
		This method will remove similar offspring in the offspring_pool, and replace any offspring with clusters in the population (if desired).

		:param population: This is the population.
		:type  population: Organisms.GA.Population.Population
		:param offspring_pool: This is all the offspring.
		:type  offspring_pool: Organisms.GA.Offspring_Pool.Offspring_Pool
		:param offspring_to_remove: These are all the offspring to remove from the offspring_pool. 
		:type  offspring_to_remove: list of int
		:param population_to_be_replaced_by_offspring: These are the names of the clusters in the population to be replaced by offspring in the offspring_pool. This is a list of (name of cluster in population to be replaced, name of offsring to replaced that cluster). 
		:type  population_to_be_replaced_by_offspring: list of (int, int)

		"""
		# Remove offspring that violate the predation Operator
		self.remove_offspring_that_violate_the_predation_operator(offspring_pool, offspring_to_remove)
		# Replace clusters in the population with offspring to comply with the predation Operator.
		self.replace_population_with_offspring(population, offspring_pool, population_to_be_replaced_by_offspring)
		#import pdb; pdb.set_trace()

	def remove_offspring_that_violate_the_predation_operator(self,offspring_pool,offspring_to_remove):
		"""
		Remove offspring from the Offspring_Pool

		:param offspring_pool: This is all the offspring.
		:type  offspring_pool: Organisms.GA.Offspring_Pool.Offspring_Pool
		:param offspring_to_remove: These are all the offspring to remove from the offspring_pool. 
		:type  offspring_to_remove: list of int

		"""
		offspring_to_remove.sort(key=lambda x:x[1],reverse=True)
		for offspring_name, offspring_index in offspring_to_remove:
			del offspring_pool[offspring_index]

	def replace_population_with_offspring(self, population, offspring_pool, population_to_be_replaced_by_offspring):
		"""
		Replace clusters in the population with offspring in the offspring_pool. These population clusters are replaced by similar offspring that have higher fitnesses than the offspring in the Offspring_Pool.
		
		:param population: This is the population.
		:type  population: Organisms.GA.Population.Population
		:param offspring_pool: This is all the offspring.
		:type  offspring_pool: Organisms.GA.Offspring_Pool.Offspring_Pool
		:param population_to_be_replaced_by_offspring: These are the names of the clusters in the population to be replaced by offspring in the offspring_pool. This is a list of (name of cluster in population to be replaced, name of offsring to replaced that cluster). 
		:type  population_to_be_replaced_by_offspring: list of (int, int)

		"""
		population_to_be_replaced_by_offspring.sort(key=lambda x:x[1],reverse=True) # not needed if the population_to_be_replaced_by_offspring list has already been sorted beforehand.
		for pop_cluster_name, offspring_name in population_to_be_replaced_by_offspring:
			pop_to_swap_out_index = population.get_index(pop_cluster_name)
			off_to_swap_in_index = offspring_pool.get_index(offspring_name)
			offspring = offspring_pool.pop(off_to_swap_in_index)
			population.replace(pop_to_swap_out_index, offspring)




import numpy as np

def get_distance(atom1,atom2):
    """
    This gives the distance between two atoms in a cluster. 

    :returns: The distance between two atoms in a cluster
    :rtype:   float

    """
    x_dist = atom1.x - atom2.x
    y_dist = atom1.y - atom2.y
    z_dist = atom1.z - atom2.z
    distance = (x_dist**2.0 + y_dist**2.0 + z_dist**2.0)**0.5
    return distance

def get_cluster_distance_list(cluster, neighbor_cutoff):
    """
    This method give a list of the interatomic distances between evety atom in the cluster. 

    :param cluster: This is the cluster to get all the interatomic distances between every atom. 
    :type  cluster: Organisms.GA.Cluster.Cluster
    :param neighbor_cutoff: If desired, this method can be programmed to not include any distances that are larger than some cutoff value. Given in Angstroms.
    :type  neighbor_cutoff: float

    :returns: A list of all the interatomic distances between every atom in the cluster. 
    :rtype:   float

    """
    list_of_distances = []
    for index_a1 in range(len(cluster)):
        for index_a2 in range(index_a1+1,len(cluster)):
            distance = get_distance(cluster[index_a1],cluster[index_a2])
            list_of_distances.append(distance)
    list_of_distances.sort()
    list_of_distances = np.array(list_of_distances)
    return list_of_distances

########################################################################################

def LoD_compare_two_structures(LoD_1, LoD_2, percentage_diff):
    """
    This method will determine if these two clusters are structurally similar or not based on the IDCM

    :param LoD_1: This is the list of the interatomic distances in cluster 1
    :type  LoD_1: list of floats
    :param LoD_2: This is the list of the interatomic distances in cluster 2
    :type  LoD_2: list of floats
    :param percentage_diff: If the differences between all the distances in LoD_1 and LoD_2 are less than this percentage difference, they are the same. If any one difference is greater than this percentage, the two clusters are different. 
    :type  percentage_diff: float

    :returns: If the differences between all the distances in LoD_1 and LoD_2 are less than percentage_diff, they are the same. If any one differences in creater than percentage_diff percent, the two clusters are different. 
    :rtype:   bool. 

    """
    similar_element_list = []
    for index in range(len(LoD_1)):
        distance1 = LoD_1[index]
        distance2 = LoD_2[index]
        distance_difference_percentage = (float(abs(distance1 - distance2))/float(min([distance1,distance2])))*100.0
        if abs(distance_difference_percentage) <= float(percentage_diff):
            pass
        else:
            return False
    return True




from Organisms.GA.Collections_Iterator import Collections_Iterator

def get_lowest_and_highest_energies_from_collections(population,collections):
	"""
	This method will return the value of the highest and lowest energy from a list of collections that are inputted into this method.
	The collections variable is a list of collections, for example [population,offspring]
	This is a private method.

	:param collections: A list of all the collections that you want to compare..
	:type  collections: list of collection objects

	:returns: the lowest energy of clusters out of all the inputed collections, the maximum energy out of all the inputed collections.
	:rtypes: float, float

	"""
	lowest_energy  =  population[0].energy
	highest_energy =  population[0].energy
	collections_iterator = Collections_Iterator(population,collections,pass_first_cluster=True)
	for cluster in collections_iterator:
		if cluster.energy < lowest_energy:
			lowest_energy = cluster.energy
		if cluster.energy > highest_energy:
			highest_energy = cluster.energy
	return lowest_energy, highest_energy

##########################################################################################################################################################

def check_rho_i(rho_i,cluster,max_energy,min_energy):
	"""
	This method will check that the rho value for the cluster is within the expected limits. If there are any errors, the program will mention what is wrong and terminate.

	:param rho_i: This is the rho value to check.
	:type  rho_i: float
	:param collection: This is the cluster that contains the associated rho value to check.
	:type  collection: Organisms.GA.Cluster
	:param min_energy: This is the lowest energy of the current lowest energetic structure across all the collections in the GA (e.g. [population, offspring]).
	:type  min_energy: float
	:param max_energy: This is the highest energy of the current highest energetic structure across all the collections in the GA (e.g. [population, offspring]).
	:type  max_energy: float
		
	"""
	if not (0.0 <= rho_i and rho_i <= 1.0):
		print('Error with def check_row_i in Energy_Diversity_Operator.py')
		print('There is no error with this method in Energy_Diversity_Operator.py, there is an error with the Diversity Operator you are using.')
		print('The value of rho must be between 0.0 and 1.0. This is not the case here.')
		print('The value of this rho_i = '+str(rho_i))
		print('Given Cluster: '+str(cluster))
		print('Given Cluster energy: '+str(cluster.energy)+' eV')
		print('Given max_energy: '+str(max_energy)+' eV')
		print('Given min_energy: '+str(min_energy)+' eV')
		print('NOTE:')
		print('Checking if min_energy < cluster.energy < max_energy:')
		print('min_energy <= cluster.energy = '+str(min_energy <= cluster.energy))
		print('cluster.energy <= max_energy = '+str(cluster.energy <= max_energy))
		print('Check this.')
		import pdb; pdb.set_trace()
		exit('This program will now end.')
		
def get_rho_i(cluster,highest_energy,lowest_energy):
	"""
	Get the rho of a cluster. Note that the max and min energy are based on that of the current population. 

	See `Theoretical study of Cu-Au nanoalloy clusters using a genetic algorithm, Darby et al., 116, 1536 (2002); doi: 10.1063/1.1429658 <http://dx.doi.org/10.1063/1.1429658>`_, page 1538 about this fitness equation, including the definition of max_energy and min_energy.

	:param cluster: This is the cluster you want to obtain rho_i for.
	:type  cluster: Cluster
	:param highest_energy: This is the highest energy of the current highest energetic structure across all the collections in the GA (e.g. [population, offspring]).
	:type  highest_energy: float
	:param lowest_energy: This is the lowest energy of the current lowest energetic structure across all the collections in the GA (e.g. [population, offspring]).
	:type  lowest_energy: float

	:returns: The value for rho for the cluster
	:rtypes: float
	"""
	energy_i = float(cluster.energy)
	rho_i = (energy_i - lowest_energy)/(highest_energy - lowest_energy)
	check_rho_i(rho_i,cluster,highest_energy,lowest_energy)
	return rho_i

def get_energetic_fitness_contribution(cluster,highest_energy,lowest_energy,energy_fitness_function):
	"""
	Get the fitness, based on the energy, for a cluster. 

	See `Theoretical study of Cu-Au nanoalloy clusters using a genetic algorithm, Darby et al., 116, 1536 (2002); doi: 10.1063/1.1429658 <http://dx.doi.org/10.1063/1.1429658>`_, page 1538 about this fitness equation, including the definition of max_energy and min_energy.

	:param cluster: This is the cluster you want to obtain rho_i for.
	:type  cluster: Cluster
	:param population: This is the population which will be used to get the max_energy and min_energy to obtain rho_i.
	:type  population: Population

	:returns: The value for energetic fitness value for the cluster
	:rtypes: float
	"""
	rho_i = get_rho_i(cluster,highest_energy,lowest_energy)
	fitness = energy_fitness_function.get_fitness(rho_i)
	return fitness

##########################################################################################################################################################




from Organisms.GA.SCM_Scripts.SCM_initialisation import get_rCuts

def CNA_Database_Check(self,fitness_information, predation_operator):

	# perform checks to make sure that the cna_database is being used appropriately between the SCM predation operator and the structure + energy fitness operator. 
	# --------------------------------------------------- #
	if ('Use Predation Information' in fitness_information):
		self.use_predation_information = fitness_information['Use Predation Information']
	else:
		self.use_predation_information = False

	have_SCM_Scheme = ('SCM Scheme' in fitness_information)

	have_rCut_high = ('rCut_high' in fitness_information)
	have_rCut_low = ('rCut_low' in fitness_information)
	have_rCut_resolution = ('rCut_resolution' in fitness_information)
	have_rCut = ('rCut' in fitness_information)
	
	have_lattice_constant = ('lattice_constant' in fitness_information)
	have_nn_high = ('nn_high' in fitness_information)
	have_nn_low = ('nn_low' in fitness_information)
	have_nn_resolution = ('nn_resolution' in fitness_information)
	have_single_nn_measurement = ('single_nn_measurement' in fitness_information)

	have_settings = (have_SCM_Scheme,have_rCut_high,have_rCut_low,have_rCut_resolution,have_rCut,have_lattice_constant,have_nn_high,have_nn_low,have_nn_resolution,have_single_nn_measurement)
	
	# ----------------------------------------------------------------------------------------------------------------------------------------
	# You are using the SCM predation operator and use_predation_information == True, so this fitness operator should use the same cna_database as the predation operator
	if   (self.predation_switch == 'SCM')     and     self.use_predation_information:
		Checks_for_if_predation_switch_equals_SCM_predation_and_use_predation_information_is_true(self,fitness_information,have_settings)
		will_make_new_cna_database = False # use the same database as the SCM predation operator
	# You are using the SCM predation operator but use_predation_information == False, so this fitness operator should use its own cna_database 
	elif (self.predation_switch == 'SCM')     and not self.use_predation_information:
		Checks_for_if_predation_switch_equals_SCM_predation_or_use_predation_information_is_true(self,fitness_information,predation_operator,have_settings)
		will_make_new_cna_database = True # You want to have a new CNA database that only the structure + energy fitness operator uses.
	# You are not using the SCM predation operator but use_predation_information == True, these two conflict with eachother. Stop the program
	elif (not self.predation_switch == 'SCM') and     self.use_predation_information:
		print('ERROR in def SCM_options, in Class SCM_and_Energy_Fitness_Operator, in SCM_and_Energy_Fitness_Operator.py/SCM_and_Energy_Fitness_Operator_Options.py')
		print("You are not using the SCM operator, however you have set the 'Use Predation Information' switch in your fitness_information dictionary as True.")
		print("You can't use the same information from the Predation operator unless the SCM predation operator is used")
		print('Check this.')
		exit('This program will finish without completing.')
	# You are not using the SCM predation operator and use_predation_information == False, so this fitness operator should use its own cna_database 
	elif (not self.predation_switch == 'SCM') and not self.use_predation_information:
		Checks_for_if_predation_switch_equals_SCM_predation_or_use_predation_information_is_true(self,fitness_information,predation_operator,have_settings)
		will_make_new_cna_database = True # You want to have a new CNA database that only the structure + energy fitness operator uses.
	else:
		print('ERROR in def SCM_options, in Class SCM_and_Energy_Fitness_Operator, in SCM_and_Energy_Fitness_Operator.py/SCM_and_Energy_Fitness_Operator_Options.py')
		print('Something weird has happened because you should not have entered into this part of the code')
		exit('Weird error. Cancelling program before it begins. Check your inputs and possibly this code.')
	return will_make_new_cna_database
	# ----------------------------------------------------------------------------------------------------------------------------------------

def Checks_for_if_predation_switch_equals_SCM_predation_and_use_predation_information_is_true(self,fitness_information,have_settings):
	have_SCM_Scheme,have_rCut_high,have_rCut_low,have_rCut_resolution,have_rCut,have_lattice_constant,have_nn_high,have_nn_low,have_nn_resolution,have_single_nn_measurement = have_settings
	if (have_SCM_Scheme or have_rCut_high or have_rCut_low or have_rCut_resolution or have_rCut or have_lattice_constant or have_nn_high or have_nn_low or have_nn_resolution or have_single_nn_measurement):
		print('ERROR in def SCM_options, in Class SCM_and_Energy_Fitness_Operator, in SCM_and_Energy_Fitness_Operator.py/SCM_and_Energy_Fitness_Operator_Options.py')
		print("You are using the SCM operator and have set the 'Use Predation Information' switch in your fitness_information dictionary to True.")
		settings_to_mention = []
		names = ['SCM Scheme','rCut_high','rCut_low','rCut_resolution','rCut','lattice_constant','nn_high','nn_low','nn_resolution','single_nn_measurement']
		for name in names:
			if name in fitness_information:
				settings_to_mention.append(name)
		print('You have also included the following settings in the fitness_information dictionary: '+' '.join(settings_to_mention))
		print('For clarity of the input settings for future readers of your input, we require that if you use the SCM predation operator and set the "Use Predation Information" switch in your fitness_information dictionary to True')
		print('then do not include any of the following settings in your input fitness_information dictionary: '+' '.join(names))
		print()
		print('Your fitness_information dictionary is: '+str(fitness_information))
		print('Check this.')
		exit('This program will finish without completing.')

def Checks_for_if_predation_switch_equals_SCM_predation_or_use_predation_information_is_true(self,fitness_information,predation_operator,have_settings):
	have_SCM_Scheme,have_rCut_high,have_rCut_low,have_rCut_resolution,have_rCut,have_lattice_constant,have_nn_high,have_nn_low,have_nn_resolution,have_single_nn_measurement = have_settings
	if not 'SCM Scheme' in fitness_information:
		print('ERROR in def SCM_options, in Class SCM_and_Energy_Fitness_Operator, in SCM_and_Energy_Fitness_Operator.py/SCM_and_Energy_Fitness_Operator_Options.py')
		print('You need to include the "SCM Scheme" setting. This could be either the T-SCM or the A-SCM.')
		print('Your fitness_information dictionary is: '+str(fitness_information))
		print('Check this.')
		exit('This program will finish without completing.')
	have_all_settings_for_rCuts = (have_rCut_high and have_rCut_low and have_rCut_resolution)
	have_all_settings_for_rCut  = (have_rCut)
	have_all_settings_for_nns   = (have_lattice_constant and have_nn_high and have_nn_low and have_nn_resolution)
	have_all_settings_for_nn    = (have_lattice_constant and have_single_nn_measurement)
	one_of_settings_for_rCuts = (have_rCut_high or have_rCut_low or have_rCut_resolution)
	one_of_settings_for_rCut  = (have_rCut)
	one_of_settings_for_nns   = (have_nn_high or have_nn_low or have_nn_resolution)
	one_of_settings_for_nn    = (have_single_nn_measurement)
	if   have_all_settings_for_rCuts:
		if (one_of_settings_for_rCut or one_of_settings_for_nns or one_of_settings_for_nn or have_lattice_constant):
			multiple_settings_error(fitness_information); exit()
	elif have_all_settings_for_rCut:
		if (one_of_settings_for_rCuts or one_of_settings_for_nns or one_of_settings_for_nn or have_lattice_constant):
			multiple_settings_error(fitness_information); exit()
	elif have_all_settings_for_nns:
		if (one_of_settings_for_rCuts or one_of_settings_for_rCut or one_of_settings_for_nn):
			multiple_settings_error(fitness_information); exit()
	elif have_all_settings_for_nn:
		if (one_of_settings_for_rCuts or one_of_settings_for_rCut or one_of_settings_for_nns):
			multiple_settings_error(fitness_information); exit()
	else:
		print('ERROR in def SCM_options, in Class SCM_and_Energy_Fitness_Operator, in SCM_and_Energy_Fitness_Operator.py/SCM_and_Energy_Fitness_Operator_Options.py')
		print('You can only include one of the following sets of settings for the fitness_information dictionary')
		print('These settings are to determine what rCut value(s) to sample with the structure + energy fitness operator.')
		print('Either one of the following set:')
		print('    -> rCut_high, rCut_low, rCut_resolution')
		print('    -> rCut')
		print('    -> lattice_constant, nn_high, nn_low, nn_resolution')
		print('    -> lattice_constant, single_nn_measurement')
		print()
		print('Your fitness_information dictionary is: '+str(fitness_information))
		print('Check this.')
		exit('This program will finish without completing.')
	# Make sure that the CNA database in the SCM predation operator would not be the same as the one that will be used in the fitness operator. 
	if (self.predation_switch == 'SCM'):
		check_rCuts_in_fitness_operator = sorted(get_rCuts(self,fitness_information))
		if check_rCuts_in_fitness_operator == sorted(predation_operator.rCuts):
			print('ERROR in def SCM_options, in Class SCM_and_Energy_Fitness_Operator, in SCM_and_Energy_Fitness_Operator.py/SCM_and_Energy_Fitness_Operator_Options.py')
			print('The values of rCut that you will use for your SCM predation method and the structure + energy fitness operator are the same')
			print('This may be because in either the predation or fitness operator you use nearest neighbour settings and in the other you use rCut settings')
			print('In any case, double check your settings for your predation and fitness operators, or if you can not find a problem potentially check the code.')
			print()
			print('Your fitness_information dictionary is: '+str(fitness_information))
			print('Your predation_operator is: '+str(predation_operator))
			print()
			print('Check this.')
			import pdb; pdb.set_trace()
			exit('This program will finish without completing.')

def multiple_settings_error(fitness_information):
	print('ERROR in def SCM_options, in Class SCM_and_Energy_Fitness_Operator, in SCM_and_Energy_Fitness_Operator.py/SCM_and_Energy_Fitness_Operator_Options.py')
	print('You have include conflicting settings in your fitness_information dictionary for the structure + energy fitness operator')
	print('You need to include only one of the following sets of settings for the fitness_information dictionary to tell the structure + energy fitness operator what rCut value(s) to sample.')
	print('Either one of the following set:')
	print('    -> rCut_high, rCut_low, rCut_resolution')
	print('    -> rCut')
	print('    -> lattice_constant, nn_high, nn_low, nn_resolution')
	print('    -> lattice_constant, single_nn_measurement')
	print()
	settings_to_mention = []
	names = ['rCut_high','rCut_low','rCut_resolution','rCut','lattice_constant','nn_high','nn_low','nn_resolution','single_nn_measurement']
	for name in names:
		if name in fitness_information:
			settings_to_mention.append(name)
	print('You have included the following settings in the fitness_information dictionary: '+' '.join(settings_to_mention))
	print()
	print('Your fitness_information dictionary is: '+str(fitness_information))
	print('Check this.')
	exit('This program will finish without completing.')




import os, sys
from time import time

from Organisms.GA.GA_Setup import GA_Setup
from Organisms.GA.GA_Introducing_Remarks import Introducing_Remarks
from Organisms.GA.GA_Initiate import GA_Initiate
from Organisms.GA.GA_Program_external_methods import if_to_finish_because_found_cluster_energy, check_names_1, check_names_2, add_metadata, floor_float, reset_population, check_files_for_readable_and_writable
from Organisms.GA.Lock import Lock_Check_and_Set #, Lock_Remove

from Organisms.GA.Get_Offspring import Create_An_Offspring

from io import StringIO
import multiprocessing as mp
from Organisms.GA.exitting_procedure import add_to_exitting_procedure

class GA_Program():
	"""
	The Otago Research Genetic Algorithm for Nanoclusters, Including Structural Methods and Similarity (Organisms) program has been designed to perform a genetic algorithm for a nanoparticle of any composition and any size. 

	It has been designed for that it can be modified by others in the Garden group after I have left. Hopefully the code is readable and can be tuned, however I have tried to note what is happening during the code and why.
	More notes can be found in the 

	:param cluster_makeup: This contains the information on the makeup of the cluster you would like to optimise for. Format is a dictionary in the form of: {element: number of that element}
	:type  cluster_makeup: {str: int, ...}
	:param pop_size: The size of the population
	:type  pop_size: int
	:param generations: The number of generations that are run
	:type  generations: int
	:param no_offspring_per_generation: The number of offspring that are created per generation
	:type  no_offspring_per_generation: int
	:param creating_offspring_mode: This indicates how the offspring are created, either via the mating method 'followed' by the mutation method, or by only perform the mating method 'or' mutation method, or (i.e. either mating and/or mutation). See manual for how to set this.
	:type  creating_offspring_mode: str.
	:param crossover_type: This is the type of crossover that you would like to use. See the manual for more information.
	:type  crossover_type: str.
	:param mutation_types: This is a list that contains all the information about the mutation methods you would like to use. 
	:type  mutation_types: list of (str., float)
	:param chance_of_mutation: This indicates the change of a mutation occuring. See the manual on specifically how this works.
	:type  chance_of_mutation: float
	:param r_ij: This is the maximum bond distance that we would expect in this cluster. See the manual for more information. 
	:type  r_ij: float
	:param vacuum_to_add_length: This is the amount of vacuum to place around the cluster.
	:type  vacuum_to_add_length: float
	:param Minimisation_Function: This is a function that determines how to locally minimise clusters. See manual for more information.
	:type  Minimisation_Function: __func__
	:param surface_details: This functionality has not been designed yet. Default: None
	:type  surface_details: None
	:param epoch_settings: This is designed to hold the information about the epoch method. 
	:type  epoch_settings: dict.
	:param cell_length: This is the length of the cubic unit cell to construct clusters in. See manual for more information. Default: 'default'
	:type  cell_length: float
	:param predation_information: This holds all the information about the predation operator. Default: {'Predation Operator':"Off"}
	:type  predation_information: dict.
	:param fitness_information: This holds all the information about the fitness operator. Default: {'Fitness Operator':"Off"}
	:type  fitness_information: dict.
	:param ga_recording_information: Default: {}
	:type  ga_recording_information: dict.
	:param force_replace_pop_clusters_with_offspring: This will tell the genetic algorithm whether to swap clusters in the populatino with offspring if the predation operator indicates they are the same but the predation operator has a better fitness value than the cluster in the population. 
	:type  force_replace_pop_clusters_with_offspring: bool.
	:param user_initialised_population_folder: This is the directory to a folder containing any custom made clusters you would like to include in the initial population. Set this to None if you do not have any initial clusters to add into the population. Default: None
	:type  user_initialised_population_folder: str. or None
	:param rounding_criteria: The number of decimal places to round the energies of clusters made during the genetic algorithm to. Default: 2
	:type  rounding_criteria: int
	:param print_details: Verbose for this algorithm.
	:type  print_details: bool.
	:param no_of_cpus: The number of cpus that the algorithm can use via multiprocessing. Default: 1
	:type  no_of_cpus: int
	:param finish_algorithm_if_found_cluster_energy: If desired, the algorithm can finish if the LES is located. This is useful to use for methods testing. The algorithm will determine that the LES is found when the genetic algorithm locates the energy of the LES. Read the manual on how to use this. Default: None
	:type  finish_algorithm_if_found_cluster_energy: dict. or None
	:param total_length_of_running_time: The total amount of time to run the genetic algorithm for. If the algorithm is still running after this time, the algorithm will safety finish. Time given in hours. None means no limit on time, Default: None.
	:type  total_length_of_running_time: int or None
	"""
	def __init__(self,cluster_makeup,pop_size,generations,no_offspring_per_generation,creating_offspring_mode,crossover_type,mutation_types,
		chance_of_mutation,r_ij,vacuum_to_add_length,Minimisation_Function,surface_details=None,epoch_settings={'use epoch': 'off'},
		cell_length='default',memory_operator_information={'perform_memory_operator':'Off'},predation_information={'Predation_Switch':"Off"},
		fitness_information={'Fitness_Switch':"Energy"},ga_recording_information={},force_replace_pop_clusters_with_offspring=True,
		user_initialised_population_folder=None,rounding_criteria=2,print_details=False,no_of_cpus=1,finish_algorithm_if_found_cluster_energy=None,
		total_length_of_running_time=None):

		# Check lock, and if ok to run set a program lock, then
		Lock_Check_and_Set()
		# Check to make sure that all files in the folder are readable and writeable
		check_files_for_readable_and_writable()
		# Set up the genetic algorithm
		GA_Setup(self,cluster_makeup,pop_size,generations,no_offspring_per_generation,creating_offspring_mode,crossover_type,mutation_types,chance_of_mutation,r_ij,vacuum_to_add_length,Minimisation_Function,surface_details,epoch_settings,cell_length,memory_operator_information,predation_information,fitness_information,ga_recording_information,force_replace_pop_clusters_with_offspring,user_initialised_population_folder,rounding_criteria,print_details,no_of_cpus,finish_algorithm_if_found_cluster_energy,total_length_of_running_time)# introduction remarks
		# print introductory notes about your genetic algorithm run
		Introducing_Remarks(self)
		# Check the GA before beginning and initiate the GA program
		GA_Initiate(self)
		# Run the main component of the GA
		self.run_GA()
		##########################################################################################################
		##########################################################################################################
		##########################################################################################################

	def run_GA(self):
		"""
		This definition is the base of the algorithm. Once the algorithm has set itself up, this will run the genetic algorithm
		"""
		print("-----------------------------")
		print("-----------------------------")
		print("Starting Genetic Algorithm")
		print("-----------------------------")
		print("-----------------------------")
		print(" ")
		finish_because_found_cluster_energy = False
		for generation_number in range(self.starting_generation,self.generations+1):
			start_time = time()
			check_names_1(self.population)
			#if if_to_finish_because_found_cluster_energy(self):
			#	finish_because_found_cluster_energy = True
			#	break
			#Create all backup files before the generation begins
			self.population.backup_files()
			self.epoch.backup()
			# -------------------------------------------------------------------------------------------------
			self.ga_program_details.start_clock()
			print('\n')
			print("----------------------------------------------------------")
			generation_wording = "RUNNING GENERATION " + str(generation_number)
			print('#'*len(generation_wording))
			print(generation_wording)
			print('#'*len(generation_wording))
			start_time = time()
			# Get all offspring for this generation
			self.previous_cluster_name = self.get_offsprings(self.previous_cluster_name,generation_number)
			all_energies_of_offpsring = self.offspring_pool.get_cluster_energies()
			check_names_2(self.population,self.offspring_pool)
			# The energy from the offspring clusters are recorded into the energyprofile. 
			self.energyprofile.add_collection(self.offspring_pool,generation_number)
			##############################################################################
			############################# MEMORY OPERATION ###############################
			# remove clusters that are too similar to clusters in the memory
			memory_operator_timer_1_start = time()
			offspring_to_remove_because_too_similar_to_a_cluster_in_memory = self.memory_operator.check_collection(self.offspring_pool)
			clusters_being_removed_by_memory_operator = self.memory_operator.remove_similar_clusters_from_offsprng_pool(self.offspring_pool,offspring_to_remove_because_too_similar_to_a_cluster_in_memory,self.print_details)
			if not len(clusters_being_removed_by_memory_operator) == 0:
				self.ga_recording_system.add_collection(clusters_being_removed_by_memory_operator, [])
			memory_operator_timer_1_end = time()
			##############################################################################
			############################# FITNESS CHECK 1 ################################
			# Assign the fitnesses of each of the offspring. This is done before the predation check 
			# as the predation check may need to use the fitnwss values to make decisions
			# Add offspring information to the fitness operator's database, if it uses one.
			fitness_timer_1_start = time()
			self.fitness_operator.add_to_database(self.offspring_pool)
			# Assign fitnesses
			self.fitness_operator.assign_all_fitnesses_before_assess_against_predation_operator(self.offspring_pool,generation_number)
			fitness_timer_1_end = time()
			##############################################################################
			########################### PREDATION OPERATION ##############################
			predation_timer_1_start = time()
			# Add offspring information to the predation operator's database, if it uses one.
			self.predation_operator.add_to_database(self.offspring_pool)
			# Identify which offspring need to be removed and potentially population that need to be replaced by offspring due to violations in predation.
			offspring_to_remove_violated_predation_operator, population_to_be_replaced_by_offspring = self.predation_operator.assess_for_violations(self.offspring_pool,self.force_replace_pop_clusters_with_offspring)
			# assign cluster information if it was placed in the population and if it was excluded by the predation method. 
			for cluster in self.offspring_pool:
				cluster.ever_in_population = (cluster.name in population_to_be_replaced_by_offspring)
				cluster.excluded_because_violates_predation_operator = (cluster.name in offspring_to_remove_violated_predation_operator)
				cluster.initial_population = False
				cluster.removed_by_memory_operator = False
			# Record clusters that were made during this generation
			self.ga_recording_system.add_collection(self.offspring_pool, offspring_to_remove_violated_predation_operator)
			# Remove offspring that violate the predation operator, and replace clusters in the population with offspring to comply with the predation operator.
			self.predation_operator.remove_offspring_and_replace_with_population_that_violate_predation_operator(self.population, self.offspring_pool, offspring_to_remove_violated_predation_operator, population_to_be_replaced_by_offspring)
			predation_timer_1_end = time()
			##############################################################################
			############################# FITNESS CHECK 2 ################################
			fitness_timer_2_start = time()
			# Remove offspring from the fitness operator's database that were removed by the predation operator during the PREDATION OPERATION.
			offspring_to_remove_violated_predation_operator = [off_name for off_name, off_index in offspring_to_remove_violated_predation_operator]
			population_to_be_replaced_by_offspring = [pop_name for pop_name, off_name in population_to_be_replaced_by_offspring]
			self.fitness_operator.remove_from_database(offspring_to_remove_violated_predation_operator+population_to_be_replaced_by_offspring)
			# Assign fitnesses to the offspring and clusters in the population before natural selection.
			self.fitness_operator.assign_all_fitnesses_after_assess_against_predation_operator(self.offspring_pool, generation_number, offspring_to_remove_violated_predation_operator)
			fitness_timer_2_end = time()
			##############################################################################
			############################# NATURAL SELECTION ##############################
			# Make a backup of the population before offspring are added to the population.
			# This has been added just in case there is a system crash, and the Organisms program needs to be restarted. 
			natural_selection_timer_start = time()
			# Perform the natural selection proceedure.
			clusters_removed_from_the_population = self.natural_selection(self.offspring_pool,generation_number)
			# clear the offspring_pool
			cleaned_offspring_names = self.offspring_pool.clean()
			natural_selection_timer_end = time()
			# Add data to the population history about the clusters in the population after this generation and their energies. 
			self.population.add_to_history_file(generation_number)
			##############################################################################
			################################# EPOCH ######################################
			###################### PREDATION and FITNESS UPDATE ##########################
			ending_timer_start = time()
			# This part of the code will perform an epoch proceedure if required.
			# Four options can occur at this point
			# 1a. If the population has converged energetically and the fitness is based completely or in part on the energy fitness function, 
			#     we will also want to reset the predation and fitness operators as well as restarting the population from fresh.
			# 1b. If the population has converged structurally and the fitness is based completely or in part on the structural component of
			#     the structure + energy fitness function, we will also want to reset the predation and fitness operators as well as 
			#     restarting the population from fresh.
			# 1c. If a epoch occurs, then we want to reset the predation and fitness operators as well as
			#     restarting the population from fresh.
			# 2.  If an epoch does not need to occur, we want to assign a fitness to each cluster in the
			#     population based on only the clusters in the population, as well as update the information 
			#     in the predation operator if needed. 
			is_energy_in_fitness_function = self.fitness_operator.fitness_switch == 'Energy' or (self.fitness_operator.fitness_switch == 'SCM + Energy' and not self.fitness_operator.SCM_fitness_contribution == 1.0)
			is_SCM_in_fitness_function = self.fitness_operator.fitness_switch == 'SCM + Energy' and not self.fitness_operator.SCM_fitness_contribution == 0.0
			if is_energy_in_fitness_function and not self.population.is_there_an_energy_range(self.rounding_criteria):
				reset_population(self,generation_number)
			elif is_SCM_in_fitness_function and not self.fitness_operator.is_there_an_similarity_range(self.similarity_rounding_criteria):
				reset_population(self,generation_number)
			elif self.epoch.should_epoch(self.population,generation_number):
				if self.epoch.first_epoch_to_change_fitness_function:
					perform_reset_population, self.fitness_operator = self.epoch.change_fitness_function(self.fitness_operator)
					if perform_reset_population:
						reset_population(self,generation_number)
				else:
					reset_population(self,generation_number)
				#reset_population(self,generation_number)
			else:
				self.fitness_operator.remove_from_database(clusters_removed_from_the_population+cleaned_offspring_names)
				self.fitness_operator.assign_all_fitnesses_after_natural_selection(generation_number)
				self.predation_operator.remove_from_database(offspring_to_remove_violated_predation_operator+population_to_be_replaced_by_offspring+clusters_removed_from_the_population+cleaned_offspring_names)
			##############################################################################
			# Save a copy of the while population at specified generations that is a complete snapshot of the algorithm at this point. It is different to what is happening with self.ga_recording_system.add_collection
			self.ga_recording_system.record_population_at_generation(self.population,generation_number)
			# Write information about performance of this generation run to ga_program_details
			self.ga_program_details.end_clock(generation_number)
			# This is a hack to get metadata into the databases for Population, Offspring_Pool, and GA_Recording_System.
			if generation_number == 1:
				add_metadata(self)
			# save the current state of the population
			self.population.current_state_file(generation_number)
			if self.print_details:
				print(self.population)
			##############################################################################
			# The generation is now complete, remove all backup files. 
			self.population.remove_backup_files()
			self.epoch.remove_backup()
			##############################################################################
			# End with the amount of time it took to run the generation.
			ending_timer_end = time()
			end_time = time()
			fitness_timer_1 = fitness_timer_1_end - fitness_timer_1_start
			memory_operator_timer_1 = memory_operator_timer_1_end - memory_operator_timer_1_start
			predation_timer_1 = predation_timer_1_end - predation_timer_1_start
			fitness_timer_2 = fitness_timer_2_end - fitness_timer_2_start
			ending_timer = ending_timer_end - ending_timer_start
			print('Time Information')
			print('fitness_timer_1   = '+str(fitness_timer_1)+' seconds')
			print('memory_operator_timer_1   = '+str(memory_operator_timer_1)+' seconds')
			print('predation_timer_1 = '+str(predation_timer_1)+' seconds')
			print('fitness_timer_2   = '+str(fitness_timer_2)+' seconds')
			print('ending_timer      = '+str(ending_timer)+' seconds')
			print('The amount of time required for generation ' + str(generation_number) + ' is ' + str(end_time - start_time)+' seconds')
			print("----------------------------------------------------------")
			if os.path.exists('finish'):
				print('I have found a file called "finish", meaning you want to stop the GA safely')
				print('The GA is stopping now. The generation is '+str(generation_number))
				print('It is possible to resume the genetic algorithm from this point.')
				print('The genetic algorithm will end')
				break
			elif self.timer.has_elapsed_time():
				print('The GA has taken longer than the user intends'+str(self.timer.total_length_of_running_time)+' hrs.')
				print('The GA is stopping now. The generation is '+str(generation_number))
				print('It is possible to resume the genetic algorithm from this point.')
				print('The genetic algorithm will end')
				break
			print('Time elapsed: '+str(self.timer.print_elapsed_time()))
			print('Date and Time: '+str(self.timer.get_time_now()))
			if if_to_finish_because_found_cluster_energy(self,all_energies_of_offpsring):
				finish_because_found_cluster_energy = True
				break
		############################################################################################################################
		if os.path.exists('finish'):
			print('Finishing by finding a "finish" file')
		elif self.timer.has_elapsed_time():
			print('Finishing by going over the elapsed time')
		elif finish_because_found_cluster_energy:
			print('Finishing by finding the energy of the desired LES cluster')
		elif generation_number == self.generations:
			print("----------------------------------------------------------")
			print("----------------------------------------------------------")
			print('The maximum generation set by the user has been reached! The genetic algorithm process has completed!')
			print('The Algorithm will finish now.')
		elif generation_number > self.generations:
			print('Error in def run_GA in class GA_Program, in GAProgram.py')
			print('The number of generations completed is more than what the user has set.')
			print('Current generation number: '+str(generation_number))
			print('Maximum Generation for this GA run: '+str(self.generations))
			print('Check this.')
			import pdb; pdb.set_trace()
			exit()
		#self.ga_recording_system.write_GA_Recording_Database_at_End_of_GA()
		#if not (os.path.exists('finish') or finish_because_found_cluster_energy):
		#	self.remove_cluster_files()
		#self.offspring_pool.close_offspring_pool()
		print("----------------------------------------------------------")
		print('Time elapsed: '+str(self.timer.print_elapsed_time()))
		print('Date and Time: '+str(self.timer.get_time_now()))
		print("----------------------------------------------------------")
		print('Finishing the Garden Group Genetic Algorithm Successfully.')
		print("----------------------------------------------------------")
		print("----------------------------------------------------------")

	#########################################################################################################################
	#########################################################################################################################

	#----------------------------------------------------------------------------#
	#                          Offspring Creation Step                           #
	#            Will create all the clusters desired per generation.            #
	#                   Has been set up for parallelisation.                     #
	#----------------------------------------------------------------------------#

	def get_offsprings(self,previous_cluster_name,generation_number):
		tasks = self.get_tasks(previous_cluster_name,generation_number)
		print('-----------------------')
		print('Making Offspring')
		#print("Made Offspring: ", end = '')
		start_time = time()
		if self.no_of_cpus == 1:
			offsprings = []
			for task in tasks:
				offspring, toString = Create_An_Offspring(task)
				offsprings.append((offspring, toString))
		else:
			with mp.Pool(processes=self.no_of_cpus) as pool: # pool = mp.Pool()
				results = pool.map_async(Create_An_Offspring, tasks)
				results.wait()
			offsprings = results.get()
		end_time = time()
		print('-----------------------')
		print('Time taken to make offspring: '+str(end_time-start_time)+' s.')
		start_time = time()
		if self.print_details:
			print('Details of created offsprings')
		for cluster, toString in offsprings:
			if not toString == '':
				print(toString)
			self.offspring_pool.add('End',cluster) # Place the initialised cluster into the population
		end_time = time()
		print('Time taken to add offspring to Offspring_Pool: '+str(end_time-start_time)+' s.')
		print('-----------------------')
		return self.offspring_pool[-1].name

	def get_tasks(self,previous_cluster_name,generation_number):
		"""
		This provides a generator that contains the inputs needed to create the offspring via parallelisation.
		"""
		def tasks(previous_cluster_name,generation_number,population,offspring_pool_name,chance_of_mutation,r_ij,cell_length,vacuum_to_add_length,creating_offspring_mode,crossover_procedure,mutation_procedure,no_offspring_per_generation,rounding_criteria,Minimisation_Function,surface,place_cluster_where,print_details):
			run_numbers = range(previous_cluster_name+1,previous_cluster_name+self.no_offspring_per_generation+1)
			for run_number in run_numbers:
				yield (run_number,generation_number,population,offspring_pool_name,chance_of_mutation,r_ij,cell_length,vacuum_to_add_length,creating_offspring_mode,crossover_procedure,mutation_procedure,no_offspring_per_generation,rounding_criteria,Minimisation_Function,surface,place_cluster_where,print_details)
		return tasks(previous_cluster_name,generation_number,self.population,self.offspring_pool_name,self.chance_of_mutation,self.r_ij,self.cell_length,self.vacuum_to_add_length,self.creating_offspring_mode,self.crossover_procedure,self.mutation_procedure,self.no_offspring_per_generation,self.rounding_criteria,self.Minimisation_Function,self.surface,self.place_cluster_where,self.print_details)

	#########################################################################################################################
	#########################################################################################################################

	#----------------------------------------------------------------------------#
	#                           Natural Selection Step                           #
	#     Replace clusters in the population with any lower energy offspring.    #
	#----------------------------------------------------------------------------#

	def natural_selection(self,offspring_pool,generation_number):
		"""
		This definition allows the genetic algorithm to perform the natural selection proceedure.

		The algorithm works as follows:

			 - The fitness of the population and the offspring are assessed.
			 - The clusters in the population are ordered from lowest to highest fitness.
			 - The offspring are ordered from highest to lowest fitness.
			 - Replace the clusters in the population with the lowest fitnesses with the offspring with the highest fitnesses.
			 - Once done, the Natural Selection Proceedure is finished.

		:param generation_number: This is the current generation of the genetic algorithm run. This mauy not be 0 if you are restarting the algorithm
		:type  generation_number: int
		:param offspring_pool: This is the offspring pool the GA will be using.
		:type  offspring_pool: Offspring

		This Natural Selection proceedure will update self.population

		"""
		if self.print_details:
			print("-----------------------------")
			print("****---------------------****")
			print("Natural Selection for Generation " + str(generation_number))
			print(self.population.information())
			print(offspring_pool.information())
		# Obtain a list of clusters in order of lowest to highest energy for clusters in the Pool
		fitness_population = []
		for index in range(len(self.population)):
			individual = self.population[index]
			fitness_population.append([individual.fitness,index,individual.name])
		fitness_population.sort(key=lambda x:x[0])
		offspring_pool.sort_by_fitness()
		# take a note of all the clusters in the population that have been removed due to being less fit than an offspring
		clusters_removed_from_the_population = []
		# swap offspring with individuals in the population.
		index_fp = 0
		if self.print_details:
			print('Offspring that replace clusters in population due to fitness.')
		while len(offspring_pool) > 0:
			if offspring_pool[0].fitness < fitness_population[index_fp][0]:
				break
			else:
				offspring = offspring_pool.pop(0)
				index_rc = fitness_population[index_fp][1]
				clusters_removed_from_the_population.append(self.population[index_rc].name)
				if self.print_details:
					print('Replacing Cluster '+str(self.population[index_rc].name)+'(fitness = '+str(floor_float(self.population[index_rc].fitness,3))+') with Offspring '+str(offspring.name)+'(fitness = '+str(floor_float(offspring.fitness,3))+')')
				self.population.replace(index_rc,offspring)
			index_fp += 1
		if self.print_details:
			print("****---------------------****")
		return clusters_removed_from_the_population

	#########################################################################################################################
	#########################################################################################################################


import os
from setuptools import setup
#from Organisms import __version__
#def get_version_number():
#  version = __version__
#  return version

def get_version_number():
  path_to_written_version = 'Organisms/__init__.py'
  with open(path_to_written_version) as initPY:
    for line in initPY:
      if line.startswith('__version__'):
        version = eval(line.rstrip().replace('__version__ = ',''))
        break
  return version

def get_long_description():
  this_directory = os.path.abspath(os.path.dirname(__file__))
  readme_file = os.path.join(this_directory, 'README.md')
  with open(readme_file, encoding='utf-8') as f:
    long_description = f.read()
  return long_description

def find_packages(root):
  packages = []
  for root, dirs, files in os.walk(root, topdown=False):
    for name in files:
      if name.endswith('.py'):
        file = os.path.relpath(os.path.join(root, name))
        dirname = os.path.dirname(file)
        rel_dirname = os.path.relpath(dirname)
        if not rel_dirname in packages:
          packages.append(rel_dirname)
  return sorted(packages)

def find_scripts():
  scripts = []
  scripts_folders = ['Subsidiary_Programs','Postprocessing_Programs','Helpful_Programs']
  for scripts_folder in scripts_folders:
    for root, dirs, files in os.walk('Organisms/'+scripts_folder, topdown=False):
      for file in files:
        if file.endswith('.py') and not 'Main' in file:
          filepath = os.path.relpath(os.path.join(root, file))
          scripts.append(filepath)
  return sorted(scripts)

setup(name='Organisms',
      packages=find_packages(root='Organisms'),
      scripts=find_scripts(),
      version=get_version_number(),
      description="This program is designed to perform a genetic algorithm global optimisation for nanoclusters.",
      long_description=get_long_description(),
      long_description_content_type='text/markdown',
      author='Geoffrey R. Weal and Dr. Anna L. Garden',
      author_email='anna.garden@otago.ac.nz',
      url = 'https://blogs.otago.ac.nz/annagarden/',
      download_url = 'https://github.com/GardenGroupUO/Organisms/archive/v'+str(get_version_number())+'.tar.gz',
      license='GNU AFFERO GENERAL PUBLIC LICENSE',
      zip_safe=False,
      keywords = ['nanoclusters', 'nanoparticles', 'clusters', 'optimisation', 'genetic_algorithm'],
      install_requires=[
          'numpy','scipy','ase>=3.21.1','asap3>=3.11.10',
          'Sphinx','sphinx-rtd-theme','sphinx-tabs','sphinxcontrib-applehelp','sphinxcontrib-devhelp',
          'sphinxcontrib-htmlhelp','sphinxcontrib-jsmath','sphinxcontrib-plantuml','sphinxcontrib-qthelp',
          'sphinxcontrib-serializinghtml','sphinxcontrib-wiki'
          ],
      classifiers=[
        'Development Status :: 3 - Alpha',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Intended Audience :: Science/Research',      # Define that your audience are developers
        'Natural Language :: English',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: Linux',
        'Operating System :: Unix',
        'Topic :: Scientific/Engineering :: Chemistry',
        'Topic :: Scientific/Engineering :: Physics',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Natural Language :: English',
        'License :: OSI Approved :: GNU Affero General Public License v3',   # Again, pick a license
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        ],
      )



#!/usr/bin/python
'''
HasCompletedUpToGeneration.py, Geoffrey Weal, 08/03/2019

This program will determine if the genetic algorithm has completed up to a certain generation. 

This program will also return the success of your genetic algorithm trials in obtaining the putative global minimum as well as the average number of minimisations required to obtain the putative global minimum.

To get the proper average number of minimisations, you want to enter your generations value to infinity. You can do this by not entering in a value of the generation. The default is set to infinity.
'''
import os, sys
import numpy as np
import scipy.stats

def Rounding_Method(number,rounding_criteria):
    """
    This definition will round the input number to the decimal place specified by rounding_criteria.
    The input number in the genetic algorithm is usually the energy of the cluster.

    :param number: The value that you would like to round (usually energy)
    :type  number: float
    :param rounding_criteria: The number of decimal places the value should be rounded to.
    :type  rounding_criteria: int
    
    """
    if rounding_criteria > 12:
        print('Error, the maximum decimal place rounding, to avoid numerical errors, is 12')
        print('rounding_criteria: '+str(rounding_criteria))
        exit('Check this out')
    number = round(number,rounding_criteria)
    return number

def mean_confidence_interval(data, confidence=0.95):
    a = 1.0 * np.array(data)
    n = len(a)
    m, se = np.mean(a), scipy.stats.sem(a)
    ci = se * scipy.stats.t.ppf((1 + confidence) / 2., n-1)
    return m, ci

class GetLESOfAllTrials_Main:

    def __init__(self,max_generation_to_survey,set_rounding):
        self.max_generation_to_survey = max_generation_to_survey
        self.set_rounding =set_rounding
        self.run()

    def run(self):
        path = os.getcwd()
        Overall_Trials_to_check = {}
        for dirpath, dirnames, filenames in os.walk(path):
            dirnames.sort()
            if any([(dirname.startswith('Trial') and dirname.replace('Trial','').isdigit()) for dirname in dirnames]):
                print('Getting LES data for '+str(dirpath))
                LES_Details = self.get_LES_Details(dirpath, dirnames, self.max_generation_to_survey, self.set_rounding)
                #import pdb; pdb.set_trace()
                toString = ''
                toString += '******************************************************************************\n'
                toString += '******************************************************************************\n'
                toString += 'This is the information of the LES of these genetic algorithm trials.\n'
                toString += 'Trial\tdir\tGen.\tEnergy (eV)\n'
                for LES_of_a_Trial in LES_Details:
                    LES_Trial, LES_clu_dir, LES_gen, LES_energy = LES_of_a_Trial
                    toString += str(LES_Trial)+'\t'+str(LES_clu_dir)+'\t'+str(LES_gen)+'\t'+str(LES_energy)+'\n'
                toString += '******************************************************************************\n'
                toString += '******************************************************************************\n'
                toString += 'Overall Details'+'\n'
                toString += 'Lowest Recorded LES: '+str(LES_Details[0][3])+' eV or equivalent.\n'
                LLES_Details = [x for x in LES_Details if x[3] == LES_Details[0][3]]
                toString += 'No of Trials that got this LES: '+str(len(LLES_Details))+' out of '+str(len(LES_Details))+'\n'
                all_gens = [x[2] for x in LLES_Details]
                average_gen, ci_gen = mean_confidence_interval(all_gens)
                toString += 'Average Generation: '+str(average_gen)+' +- '+str(ci_gen)+'\n'
                all_dirs = [x[1] for x in LLES_Details]
                average_dir, ci_dir = mean_confidence_interval(all_dirs)
                percentage_of_successful_trials = (float(len(LLES_Details))/float(len(LES_Details)))*100.0
                percentage_of_successful_trials = round(percentage_of_successful_trials,1)
                toString += 'Average Number of Minimisations (of the '+str(len(LLES_Details))+' successful trials ('+str(percentage_of_successful_trials)+' %)): '+str(average_dir)+' +- '+str(ci_dir)+'\n'
                print(toString)
                toString += '******************************************************************************\n'
                toString += '******************************************************************************\n'
                with open(dirpath+'/LESOfTrials_'+str(self.max_generation_to_survey)+'.txt','w') as LESOfTrials:
                    LESOfTrials.write(toString)
                name = dirpath.replace(path+'/','')
                name = name.replace('/','_')
                with open(path+'/'+ name+'_LESOfTrials'+str(self.max_generation_to_survey)+'.txt','w') as LESOfTrials:
                    LESOfTrials.write(toString)
                dirnames[:] = []
                filenames[:] = []
                self.make_file_of_data(dirpath,None,LES_Details)
                self.make_file_of_data(path,name,LES_Details)

        print('########################################################################')
        print('########################################################################')
        
        print('########################################################################')
        print('########################################################################')

    def get_LES_Details(self, dirpath, dirnames, max_generation_to_survey, rounding):

        def Lowest_Energy_from_Trial(filepath, rounding):
            LES_clu_dir = -1; LES_gen = -1; LES_energy = float('inf')
            if os.path.exists(filepath):
                got_to_trial = False
                pop_size = 0
                no_of_off_per_gen = 0
                last_gen_counter = 0
                with open(filepath,'r') as DETAILS_FILE:
                    for line in DETAILS_FILE:
                        if line.startswith('Genetic Algorithm Starts Here.'):
                            continue
                        if line.startswith('Restarting due to epoch.'):
                            continue
                        if line.startswith('Finished prematurely as LES energy found.'):
                            got_to_trial = True
                            break
                        clu_dir, gen, energy = line.rstrip().split()
                        clu_dir = int(clu_dir); gen = int(gen)
                        if gen == 0:
                            pop_size += 1
                        if gen == 1:
                            no_of_off_per_gen += 1
                        if gen == max_generation_to_survey:
                            got_to_trial = True
                            last_gen_counter += 1
                        if gen > max_generation_to_survey:
                            break
                        energy = Rounding_Method(float(energy),rounding)
                        if energy < LES_energy:
                            LES_clu_dir = clu_dir; LES_gen = gen; LES_energy = energy
                if got_to_trial == False:
                    print(filepath)
                    print('Error, 2. Gen: '+ str(gen))
                    import pdb; pdb.set_trace()
                    exit()
                if gen == max_generation_to_survey:
                    if last_gen_counter < no_of_off_per_gen:
                        exit('Error, 3')
            return LES_clu_dir, LES_gen, LES_energy

        LES_Data_from_Trials = []
        dirnames = [dirname for dirname in dirnames if dirname.startswith('Trial')]
        dirnames.sort(key=lambda dirname: int(dirname.replace('Trial','')))
        number_of_dirnames = len(dirnames)
        for index in range(number_of_dirnames):
            dirname = dirnames[index]
            trial_no = int(dirname.replace('Trial',''))
            LES_clu_dir, LES_gen, LES_energy = Lowest_Energy_from_Trial(dirpath+'/'+dirname+'/Population/EnergyProfile.txt',rounding)
            LES_Data_from_Trials.append([dirname, LES_clu_dir, LES_gen, LES_energy])
            ###########
            sys.stdout.write("\r                                                                  ")
            sys.stdout.flush()
            sys.stdout.write("\rScanning Completion: "+str(float(index+1)/float(number_of_dirnames)*100.0)+" % (Checked "+str(dirname)+").")
            sys.stdout.flush()
            ###########
        LES_Data_from_Trials.sort(key=lambda x:x[3],reverse=False)
        return LES_Data_from_Trials

    def make_file_of_data(self, dirpath, name, LES_Details):

        if not name in ['', None]:
            name += '_'

        success_at_gens = [0]*(self.max_generation_to_survey+1)
        LES_Details.sort(key=lambda x:x[3],reverse=False)
        putative_LES_energy = LES_Details[0][3]
        for dirname, LES_clu_dir, LES_gen, LES_energy in LES_Details:
            if LES_energy == putative_LES_energy:
                success_at_gens[LES_gen] += 1

        with open(dirpath+'/'+str(name)+'first_instance_up_to_'+str(self.max_generation_to_survey)+'.txt','w') as first_instanceTXT:
            generation = 0
            for success_at_gen in success_at_gens:
                first_instanceTXT.write(str(generation)+': '+str(success_at_gen)+'\n')
                generation += 1

        success_at_gens_collaborated = [0]
        for success_at_gen in success_at_gens:
            at_gen_success = success_at_gens_collaborated[-1] + success_at_gen
            success_at_gens_collaborated.append(at_gen_success)
        del success_at_gens_collaborated[0]

        with open(dirpath+'/'+str(name)+'success_up_to_'+str(self.max_generation_to_survey)+'.txt','w') as successTXT:
            generation = 0
            for success_at_gen_collaborated in success_at_gens_collaborated:
                successTXT.write(str(generation)+': '+str(success_at_gen_collaborated)+'\n')
                generation += 1



'''
IDCM_Predation_Operator.py, Geoffrey Weal, 25/12/2019

This is one of the Predation Operators that can be used by the genetic algorithm program. This is the IDCM Predation Operator.

This Operator works by preventing the population from having clusters with the same ...

'''

##################################################################################################################################################
#################### Initialising and Preparing the imports and internal defs that are needed for the IDCM Predation Operator #####################
##################################################################################################################################################

import os
import numpy as np
from copy import deepcopy
from io import StringIO

from Organisms.GA.Predation_Operators.Predation_Operator import Predation_Operator

from Organisms.GA.Predation_Operators.IDCM_Predation_Operator_Scripts.IDCM_Methods import get_cluster_distance_list, LoD_compare_two_structures
from Organisms.GA.Predation_Operators.IDCM_Predation_Operator_Scripts.LoD_Comparison_Database import LoD_Comparison_Database

import multiprocessing as mp
import time
from collections import OrderedDict

##################################################################################################################################################

class Collection_and_Similarity_Settings_Iterator:
	def __init__(self, collection, similarity_settings):
		self.index = 0
		self.collection = collection
		self.similarity_settings = similarity_settings

	def __iter__(self):
		return self

	def __next__(self):
		if self.index < len(self.collection):
			index = self.index
			self.index += 1
			cluster = self.collection[index]
			return (cluster, self.similarity_settings)
		else:
			raise StopIteration

class structures_in_LoD_database_Iterator:
	def __init__(self, LoD_database, LoD_comparison_database, similarity_settings):
		self.index1 = 0
		self.index2 = self.index1+1
		self.LoD_database = LoD_database
		self.LoD_comparison_database = LoD_comparison_database
		self.structures_in_LoD_database = list(self.LoD_database.keys())
		self.similarity_settings = similarity_settings

	def __iter__(self):
		return self

	def __next__(self):
		while True:
			if self.index2 == len(self.structures_in_LoD_database):
				self.index1 += 1
				self.index2 = self.index1+1
			if self.index1+1 == len(self.structures_in_LoD_database):
				raise StopIteration
			cluster1_name = self.structures_in_LoD_database[self.index1]
			cluster2_name = self.structures_in_LoD_database[self.index2]
			self.index2 += 1
			if not self.LoD_comparison_database.is_cluster_pair_in_the_database(cluster1_name,cluster2_name):
				task_input = (cluster1_name, cluster2_name, self.LoD_database[cluster1_name], self.LoD_database[cluster2_name], self.similarity_settings)
				break
		return task_input

##################################################################################################################################################

def get_LoD_database(cluster_info):
	cluster, similarity_settings = cluster_info
	percentage_diff = similarity_settings['percentage_diff']
	cluster_LoD = get_cluster_distance_list(cluster, percentage_diff)
	return (cluster.name, cluster_LoD)

def check_if_two_clusters_are_identical_LoD(both_clusters):
	cluster_1_name, cluster_2_name, cluster_1, cluster_2, similarity_settings = both_clusters
	#print((cluster_1_name, cluster_2_name, eps_r))
	#print('Comparing '+str(cluster_1.name)+' with '+str(cluster_2.name))
	percentage_diff = similarity_settings['percentage_diff']
	the_same_cluster = LoD_compare_two_structures(cluster_1, cluster_2, percentage_diff)
	return ((cluster_1_name, cluster_2_name), the_same_cluster)

def remove_duplicates(duplicate): 
    final_list = [] 
    for num in duplicate: 
        if num not in final_list: 
            final_list.append(num) 
    return final_list 

##################################################################################################################################################

class Cluster_Block:
	"""
	This is used by the Remove_Cluster_Due_To_Predation_Violation definition to store information in an easy way to help the user understand what is going on in this method 

	:param collection: This is the collection to record. This is either the instance of the Population or the Offspring_Pool
	:type  collection: Population or Offspring_Pool
	:param collection_type: This describes if the cluster recorded is in the population (given as 'pop') or the offspring (given as 'off').
	:type  collection_type: str.
	:param index: The position/index of the cluster in the collection.
	:type  index: int

	"""
	def __init__(self,collection,collection_type,index):
		self.collection_type = collection_type
		self.index = index
		self.name = collection[index].name
		self.fitness = collection[index].fitness
		self.energy = collection[index].energy
	def  __repr__(self):
		return str(self.collection_type)+str(self.name)+'('+str(self.index)+','+str(round(self.energy,6))+' eV,'+str(round(self.fitness,3))+')'

##################################################################################################################################################

class IDCM_Predation_Operator(Predation_Operator):
	"""
	This is a Predation Operator for identifying if two clusters are identifcal using a style of a euclidean distance matrix.

	This is based on the method given in MEGA to describe the IDCM Operator.

	MEGA is described in more detail in https://pubs.acs.org/doi/abs/10.1021/acs.jpcc.6b12848

	:param Predation_Information: This contains all the information needed by the Predation Operator you want to use to run.
	:type  Predation_Information: dict.
	:param population: This is the population that this Operator will be controlling to make sure that no two clusters in the population have the same energy.
	:type  population: Organisms.GA.Population
	:param print_details: Print details of the predation operator, like verbose
	:type  print_details: bool

	"""
	def __init__(self,Predation_Information,population,no_of_cpus,print_details):
		super().__init__(Predation_Information,population,print_details)
		self.no_of_cpus = no_of_cpus 
		########################################################################################################################
		if not Predation_Information['Predation Operator'] == 'IDCM':
			print('Error in class IDCM_Predation_Operator, in IDCM_Predation_Operator.py')
			print('The Predation Operator "IDCM_Predation_Operator" has been initialised.')
			print("However, Predation_Information['Predation Operator'] is not 'IDCM'")
			print("Predation_Information['Predation Operator'] = "+str(Predation_Information['Predation Operator']))
			print('Check this out.')
			import pdb; pdb.set_trace()
			exit()
		############################################################
		self.structurally_identical_method = 'IDCM'
		if self.structurally_identical_method == 'IDCM':
			self.get_structural_data_method = get_LoD_database
			self.check_if_two_clusters_are_identical_method = check_if_two_clusters_are_identical_LoD
			percentage_diff = float(Predation_Information['percentage_diff']) # distance_difference
			neighbor_cutoff = float('inf') # Predation_Information['neighbor_cutoff'] # neighbor_cutoff
			self.similarity_settings = {'percentage_diff': percentage_diff}
		else:
			print('Error in IDCM_Predation_Operator, in IDCM_Predation_Operator.py')
			print('The structurally identical method that can be used can be:')
			print('\t- IDCM - A structural identical method that have been taken from the mexican enchanced genetic algorithm. https://doi.org/10.1021/acs.jpcc.6b12848')
			import pdb; pdb.set_trace()
			exit()
		# Create the databases that will hold the EDM information of all the clsuters in the population and offspring, and how they compare.
		self.LoD_database = {}
		self.LoD_comparison_database = LoD_Comparison_Database()
		####################################################################################

	################################################################################################################################################################################################################
	########################## Methods Required for the def Check_Initial_Population, in the class Predation_Operator, in IDCM_Predation_Operator.py ########################################################################
	################################################################################################################################################################################################################

	def check_initial_population(self,return_report=False):
		"""
		This definition is responsible for making sure that the initialised population obeys the Predation Operator of interest.

		:param return_report: Will return a dict with all the information about what clusters are similar to what other clusters in the population. 
		:type  return_report: bool.

		returns:
			* **clusters_to_remove** (*list of ints*): a list of the clusters to remove from the population as they violate the Predation Operator. Format is [(index in population, name fo cluster),...]
			* **CNA_report** (*dict.*): a dictionary with information on the clusters being removed and the other clusters in the population which have caused the violation to the SCM Predation Operator. This information is only used to display information so they know why there are violations to the Predation Operator when they occur. For is {removed cluster: [list of clusters that this cluster is similar to in the population.]}
		
		"""
		if self.print_details:
			print('--------------------------------------')
			print('Checking Population of Same Structure')
			print('--------------------------------------')
			print('Update LoD database')
		self.update_LoD_database(self.population)
		if self.print_details:
			print('Update LoD database comparison')
		self.get_identical_structures_initial_population()
		if self.print_details:
			print('Get which clusters to remove from population due to being identical')
		clusters_to_remove, identical_structures = self.pop_identical_structures()		
		if self.print_details:
			print('Removing similar clusters in the Population.')
		self.remove_similar_clusters_in_population(clusters_to_remove)
		if self.print_details and not clusters_to_remove == []:
			print('Clusters in the Population Removed for being to similar: '+str([cluster_to_remove[1] for cluster_to_remove in clusters_to_remove]))
		self.check_database([self.population])#, clusters_to_remove)
		if return_report:
			return clusters_to_remove, identical_structures
		else:
			return clusters_to_remove

	def update_LoD_database(self, collection):
		"""
		This methos will add the collection to the LoD_database.

		:param collection: This is the collection to add to the LoD_database
		:type  collection: Organisms.GA.Population.Population or Organisms.GA.Offspring_Pool.Offspring_Pool

		"""
		atoms_format_clusters = Collection_and_Similarity_Settings_Iterator(collection, self.similarity_settings)
		start_time = time.time()
		with mp.Pool(processes=self.no_of_cpus) as pool:
			results = pool.map_async(self.get_structural_data_method, atoms_format_clusters)
			results.wait()
		data = results.get()
		end_time = time.time()
		print('Collecting Cluster LoD data took '+str(end_time - start_time)+' seconds.')
		for cluster_name, cluster_EDM in data:
			self.LoD_database[cluster_name] = cluster_EDM

	def get_identical_structures_initial_population(self):
		"""
		This method is designed to identify which clusters in the LoD_database (from the initial population) are identify using this method.

		This will place all the results from this into self.LoD_comparison_database 

		"""
		# Get a list of all the comparisions to make
		structures_to_compare = structures_in_LoD_database_Iterator(self.LoD_database, self.LoD_comparison_database, self.similarity_settings)
		# Perform all the comparisons in parallel
		start_time = time.time()
		with mp.Pool(processes=self.no_of_cpus) as pool:
			results = pool.map_async(self.check_if_two_clusters_are_identical_method, structures_to_compare)
			results.wait()
		data = results.get()
		end_time = time.time()
		print('Processing Cluster LoD comparisons Entries took '+str(end_time - start_time)+' seconds.')
		# Add the comparison information to self.LoD_comparison_database. 
		for compared_cluster_names, is_the_same_cluster in data:
			name1, name2 = compared_cluster_names
			self.LoD_comparison_database.add(name1, name2, is_the_same_cluster)

	def pop_identical_structures(self):
		"""
		This method will identify which clusters need to be removed from self.LoD_comparison_database based on being identical via this method, and will remove the entries of those clusters from self.LoD_database and self.LoD_comparison_database  
		
		returns:
			* **clusters_to_remove** (*list of ints*): a list of the clusters to remove from the population as they violate the Predation Operator. Format is [(index in population, name fo cluster),...]
			* **identical_structures_report** (*dict.*): a dictionary with information on the clusters being removed and the other clusters in the population which have caused the violation to the SCM Predation Operator. This information is only used to display information so they know why there are violations to the Predation Operator when they occur. For is {removed cluster: [list of clusters that this cluster is similar to in the population.]}
		
		"""
		######################################################################
		# Get the pairs of clusters that are indentical using this Predation Operator.
		identical_pairs = self.LoD_comparison_database.which_clusters_in_LoD_comparison_database_are_similar()
		#identical_pairs_copy = deepcopy(identical_pairs)
		list_of_identical_clusters = identical_pairs.keys()
		######################################################################
		# Get a list of all the clusters in the LoD_database and order then by fitness
		list_of_identical_clusters_ordered_by_fitness = []
		for cluster_name in list_of_identical_clusters:
			cluster = self.population.get_cluster_from_name(cluster_name)
			cluster_fitness = cluster.fitness
			list_of_identical_clusters_ordered_by_fitness.append((cluster_name,cluster_fitness))
		# order fitnesses,From highest to lowest
		list_of_identical_clusters_ordered_by_fitness.sort(key = lambda x: x[1], reverse=True) 
		if (not list_of_identical_clusters_ordered_by_fitness == []) and (not list_of_identical_clusters_ordered_by_fitness[0][1] >= list_of_identical_clusters_ordered_by_fitness[-1][1]):
			print('Error in def pop_identical_structures in class CNA_Diversity_Scheme, in Energy_Diversity_Scheme.py')
			print('The fitnesses after the sort are the wrong way around')
			print('list_of_identical_clusters_ordered_by_fitness[0][1] (fitness) = '+str(list_of_identical_clusters_ordered_by_fitness[0][1]))
			print('list_of_identical_clusters_ordered_by_fitness[-1][1] (fitness)= '+str(list_of_identical_clusters_ordered_by_fitness[-1][1]))
			print('The fitnesses need to be order from highest to lowest.')
			exit('Check this out. This program will finish without completing.')
		for index in range(len(list_of_identical_clusters_ordered_by_fitness)):
			cluster_name = list_of_identical_clusters_ordered_by_fitness[index][0]
			list_of_identical_clusters_ordered_by_fitness[index] = cluster_name
		######################################################################
		# This will determine which clusters should be deleted in such a way that the 
		# high fitness clusters are kepted, and the lower fitness clusters are removed.
		# It will also remove those lower fitness identical clusters from self.LoD_database
		clusters_to_remove = []
		identical_structures_report = {}
		index1 = 0
		while index1 < len(list_of_identical_clusters_ordered_by_fitness):
			clusters_to_keep_high_fitness = list_of_identical_clusters_ordered_by_fitness[index1]
			clusters_to_remove_low_fitness = list(identical_pairs[clusters_to_keep_high_fitness])
			for cluster_to_remove in clusters_to_remove_low_fitness:
				identical_structures_report.setdefault(clusters_to_keep_high_fitness,[]).append(cluster_to_remove)
				del self.LoD_database[cluster_to_remove]
				self.LoD_comparison_database.remove(cluster_to_remove)
				# --------------------------------------------- #
				del identical_pairs[cluster_to_remove]
				for key, value in identical_pairs.items():
					if cluster_to_remove in value:
						value.remove(cluster_to_remove)
						#if len(value) == 0:
						#	del identical_pairs[key]
				# --------------------------------------------- #
				clusters_to_remove.append(cluster_to_remove)
				list_of_identical_clusters_ordered_by_fitness.remove(cluster_to_remove)
			index1 += 1
		######################################################################
		return clusters_to_remove, identical_structures_report
		######################################################################

	def remove_similar_clusters_in_population(self,clusters_to_remove):
		"""
		This method will remove the similar clusters from the population.

		:param cluster_names_to_remove: A list of the names of all the clusters to remove from the CNA database
		:type  cluster_names_to_remove: list of ints

		"""
		for index in range(len(clusters_to_remove)):
			cluster_name = clusters_to_remove[index]
			index_pop = self.population.get_index(cluster_name)
			clusters_to_remove[index] = (index_pop,cluster_name)
		clusters_to_remove.sort(key = lambda cluster_to_remove: cluster_to_remove[0], reverse=False) # sort from lowest to highest index
		for cluster_to_remove in clusters_to_remove[::-1]: # remove from highest to lowest index (by doing clusters_to_remove[::-1])
			index_pop = cluster_to_remove[0]
			self.population.remove(index_pop)

	def check_database(self, collections):
		"""
		This method will check to make sure that all the clusters comparisons in the self.LoD_comparison_database are false

		I.e. check that all the clusters in the collections are not idenitcal

		:param collections: This is a list of all the Populations and Offspring_Pool in your genetic algorithm to check.
		:type  collections:  list of Organisms.GA.Collection.Collection

		"""
		total_length = 0
		for collection in collections:
			total_length += len(collection)
		#############################################
		if not total_length == len(self.LoD_database):
			print('Error in def Check_Initial_Population in Class IDCM_Predation_Operator, in IDCM_Predation_Operator.py')
			print('not len(collections) == len(self.EDM_database)')
			print('collections: '+str(collections))
			print('')
			for collection in collections:
				for cluster in collection:
					print(cluster.name)
				print('')
			print('')
			print('total_length = '+str(total_length))
			print('len(self.LoD_database)) = '+str(len(self.LoD_database)))
			print('Check this')
			import pdb; pdb.set_trace()
			exit('This program will exit without completing')
		#############################################
		self.LoD_comparison_database.are_all_entries_false([self.population])

	###########################################################################################################################################################
	########################## Methods Required for the def assess_for_violations, in the class Fitness_Factor, in Fitness_Factor.py ##########################
	###########################################################################################################################################################

	def assess_for_violations(self,offspring_pool,force_replace_pop_clusters_with_offspring):
		"""
		The offspring are assessed against clusters in the population. Offspring are removed from the offspring_pool if:
		1) They are geometrically the same as another offspring
		2) They are geometrically the same as another cluster in the population.

		Must return the clusters that have been removed from population and offspring_pool lists.

		:param offspring_pool: This is the collection of offspring to assess for violations to the Predation Operator.
		:type  offspring_pool: Organisms.GA.Offspring_Pool.Offspring_Pool
		:param force_replace_pop_clusters_with_offspring: This will tell the genetic algorithm whether to swap clusters in the population with offspring if the predation operator indicates they are the same but the predation operator has a better fitness value than the cluster in the population. 
		:type  force_replace_pop_clusters_with_offspring: bool.

		returns:
			* **offspring_to_remove** (*tuple of ints*): A list of the names of the offspring to be removed
			* **force_replacement** (*tuple of (int, int)*): A list of the clusters in the population that should be replaced, and the offspring they should be replaced by.

		"""
		if force_replace_pop_clusters_with_offspring:
			return self.assess_for_violations_force_replacement(offspring_pool)
		else:
			return self.assess_for_violations_no_force_replacement(offspring_pool)

	def assess_for_violations_no_force_replacement(self,offspring_pool):
		"""
		The offspring are assessed against clusters in the population. Offspring are removed from the offspring_pool if:
		1) They are geometrically the same as another offspring
		2) They are geometrically the same as another cluster in the population.

		Must return the clusters that have been removed from population and offspring_pool lists.

		:param offspring_pool: This is the collection of offspring to assess for violations to the Predation Operator.
		:type  offspring_pool: Organisms.GA.Offspring_Pool.Offspring_Pool

		returns:
			* **offspring_to_remove** (*tuple of ints*): A list of the names of the offspring to be removed
			* **force_replacement** (*tuple of (int, int)*): A list of the clusters in the population that should be replaced, and the offspring they should be replaced by.

		"""
		if self.print_details:
			print('**********************************************************')
			print('**********************************************************')
			print('Offspring removed from the Offspring Pool due to violating the Comprehensive Energy Diversity Scheme.')
			print('NOTE: energy difference minimum = '+str(self.minimum_energy_diff)+' eV.')
			print('**********************************************************')
			print('**********************************************************')
		########################################################################################################
		if self.print_details:
			print('Start assessing for IDCM Diversity')
		offspring_to_remove = []
		# Check offspring against themselves
		structures_to_consider = [Cluster_Block(offspring_pool,'Offspring',index) for index in range(len(offspring_pool))]
		structures_to_consider.sort(key=lambda cluster: cluster.fitness, reverse=False)
		index1 = 0
		while index1 < len(structures_to_consider):
			offspring1 = structures_to_consider[index1]
			for index2 in range(len(structures_to_consider)-1,index1,-1):
				offspring2 = structures_to_consider[index2]
				if self.LoD_comparison_database.LoD_Similarity_Analysis(offspring1.name,offspring2.name):
					offspring_to_remove.append((offspring2.name,offspring2.index))
					del structures_to_consider[index2]
			index1 += 1
		# check offspring against clusters in the populatino
		for index in range(len(structures_to_consider)):
			offspring = structures_to_consider[index]
			for cluster_in_population in self.population:
				if self.LoD_comparison_database.LoD_Similarity_Analysis(offspring.name,cluster_in_population.name):
					offspring_to_remove.append((offspring.name,offspring.index))
					break
		# Sort by index and continue on
		offspring_to_remove.sort(key=lambda x:x[1],reverse=True)
		force_replacement = []
		return offspring_to_remove, force_replacement

	def assess_for_violations_force_replacement(self,offspring_pool):
		"""
		The offspring are assessed against clusters in the population. Offspring are removed from the offspring_pool if:
		1) They are geometrically the same as another offspring
		2) They are geometrically the same as another cluster in the population.

		Must return the clusters that have been removed from population and offspring_pool lists.

		:param offspring_pool: This is the collection of offspring to assess for violations to the Predation Operator.
		:type  offspring_pool: Organisms.GA.Offspring_Pool.Offspring_Pool

		returns:
			* **offspring_to_remove** (*tuple of ints*): A list of the names of the offspring to be removed
			* **force_replacement** (*tuple of (int, int)*): A list of the clusters in the population that should be replaced, and the offspring they should be replaced by.

		"""
		if self.print_details:
			print('------------------------------------------------------')
			print('Assessing the clusters in the offspring to decide if:')
			print('\t* offspring are geometrically similar to other offspring.')
			print('\t* offspring are geometrically similar to clusters in the population.')
			print('------------------------------------------------------')
			####################################################################################################
			print('Starting Sorting pop and off by fitness')
		all_clusters_sorted_by_expected_fitness = [Cluster_Block(self.population,'Population',index) for index in range(len(self.population))] + [Cluster_Block(offspring_pool,'Offspring',index) for index in range(len(offspring_pool))]
		# order fitnesses,From highest to lowest
		all_clusters_sorted_by_expected_fitness.sort(key=lambda cluster: cluster.fitness,reverse=True)
		if (not all_clusters_sorted_by_expected_fitness == []) and (not all_clusters_sorted_by_expected_fitness[0].fitness >= all_clusters_sorted_by_expected_fitness[-1].fitness):
			print('Error in def assess_for_violations in class CNA_Diversity_Scheme, in Energy_Diversity_Scheme.py')
			print('The fitnesses after the sort are the wrong way around')
			print('structures_to_consider[0].fitness = '+str(structures_to_consider[0].fitness))
			print('structures_to_consider[-1].fitness = '+str(structures_to_consider[-1].fitness))
			print('The fitnesses need to be order from highest to lowest.')
			exit('Check this out. This program will finish without completing.')
		if self.print_details:
			print('Finishing Sorting pop and off by fitness')
			####################################################################################################
			# Follows same rough algorithm guide as the comprehensive energy Predation Operator
			print('Start assessing for LoD Predation')
		index_higher = 0
		swap_cluster_in_pop_with_offspring = []; offspring_to_remove_due_to_pop = []; offspring_to_remove_due_to_other_offspring = []
		offspring_to_remove_as_too_many_pop_in_similar_common = []
		# scan though all the clusters in offspring and population in order of fitness
		while index_higher < len(all_clusters_sorted_by_expected_fitness):
			################################################################################################################
			incountered_population_counter = 0
			cluster_higher = all_clusters_sorted_by_expected_fitness[index_higher]
			one_instance_of_swap_cluster_in_pop_with_offspring = []
			offspring_to_remove_due_to_pop_temp = []
			offspring_to_remove_as_too_many_pop_in_similar_common_temp = []
			offspring_to_remove_due_to_other_offspring_temp = []
			# scan though all the other clusters which have higher fitnesses
			indices_to_delete = []
			for index_lower in range(index_higher+1, len(all_clusters_sorted_by_expected_fitness)):
				cluster_lower = all_clusters_sorted_by_expected_fitness[index_lower]
				############################################################################################################
				if self.LoD_comparison_database.LoD_Similarity_Analysis(cluster_higher.name,cluster_lower.name):   
					# The two clusters being looked at at the moment are geometrically the same. 
					if cluster_higher.collection_type == 'Population':
						# if two clusters in the population are geometrically similar, we have an issue.
						if cluster_lower.collection_type == 'Population':
							print('Error')
							import pdb; pdb.set_trace()
							exit()
						# If a cluster in the population (and has a higher fitness) is geometrically similar to an offspring,
						# then it is better to keep the offspring and remove the offspring
						elif cluster_lower.collection_type == 'Offspring':
							offspring_to_remove_due_to_pop_temp.append([cluster_lower,cluster_higher])
					elif cluster_higher.collection_type == 'Offspring':
						# If an offspring is geometrically similar (and has a higher fitness) to a cluster in population,
						# then it is better to replace this cluster in the population with this offspring.
						if cluster_lower.collection_type == 'Population':
							# Take the lowest fitness cluster in the population to swap with this offspring.
							if one_instance_of_swap_cluster_in_pop_with_offspring == []:
								one_instance_of_swap_cluster_in_pop_with_offspring = [[cluster_lower,cluster_higher]]
							else:
								cluster_lower_old, cluster_higher_old = one_instance_of_swap_cluster_in_pop_with_offspring[0]
								if not cluster_higher_old.name == cluster_higher.name:
									exit('Somethings up')
								offspring_to_remove_as_too_many_pop_in_similar_common_temp.append([cluster_higher,[cluster_lower_old,cluster_lower]])
								one_instance_of_swap_cluster_in_pop_with_offspring = []; indices_to_delete = []
								break
						# If an offspring is geometrically similar (and has a higher fitness) to an offspring,
						# remove the higher fitness offspring
						elif cluster_lower.collection_type == 'Offspring':
							offspring_to_remove_due_to_other_offspring_temp.append([cluster_lower,cluster_higher])
					indices_to_delete.append(index_lower)
			################################################################################################################
			if not offspring_to_remove_as_too_many_pop_in_similar_common_temp == []:
				offspring_to_remove_as_too_many_pop_in_similar_common += offspring_to_remove_as_too_many_pop_in_similar_common_temp
				del all_clusters_sorted_by_expected_fitness[index_higher]
			else:
				offspring_to_remove_due_to_pop += offspring_to_remove_due_to_pop_temp
				offspring_to_remove_due_to_other_offspring += offspring_to_remove_due_to_other_offspring_temp
				swap_cluster_in_pop_with_offspring += one_instance_of_swap_cluster_in_pop_with_offspring
				for index_lower in sorted(indices_to_delete,reverse=True): # remove from highest to lowest index
					del all_clusters_sorted_by_expected_fitness[index_lower]
				index_higher += 1
			################################################################################################################
		########################################################################################################
		########################################################################################################
		# remove offspring that violate the comprehensive Predation method as it has a higher energy, but within 
		# self.minimum_energy_diff, with other offspring or other clusters in the population.
		if self.print_details:
			print('Finished assessing for LoD Predation')
			print('Start deleting offspring')
		all_clusters_to_be_removed_here = offspring_to_remove_due_to_pop+offspring_to_remove_due_to_other_offspring+offspring_to_remove_as_too_many_pop_in_similar_common
		offspring_to_remove = []
		for off_to_remove, pop_to_keep in sorted(all_clusters_to_be_removed_here,key=lambda x:x[0].index,reverse=True): # remove from highest to lowest index
			if offspring_pool[off_to_remove.index].name in [offspring.name for cluster,offspring in swap_cluster_in_pop_with_offspring]:
				print('Error in def assess_for_violations, in class IDCM_Predation_Operator, in IDCM_Predation_Operator.py')
				print('Somehow, this algorithm has identified a structure to be in the offspring_pool and the population')
				print('The structure can only be in either the offspring pool or population.')
				print('Offspring: '+str(offspring_pool[off_to_remove.index].name))
				print('Check this out')
				import pdb; pdb.set_trace()
				exit('This program will end without completing.')
			off_index = off_to_remove.index
			off_name  = offspring_pool[off_index].name
			offspring_to_remove.append((off_name,off_index))
			###############################################################################
			#name_to_remove = offspring_pool[off_to_remove.index].name
			#del self.LoD_database[name_to_remove]
			#self.LoD_comparison_database.remove(name_to_remove)
			###############################################################################
		###############################################################################
		if self.print_details:
			self.assess_for_violations_message(swap_cluster_in_pop_with_offspring,offspring_to_remove_as_too_many_pop_in_similar_common,offspring_to_remove_due_to_pop,offspring_to_remove_due_to_other_offspring)
		###############################################################################
		# Deal with the offspring that you would like to swap with clusters in the population.
		force_replacement = []
		#print(swap_cluster_in_pop_with_offspring)
		for pop_to_swap_out, off_to_swap_in in swap_cluster_in_pop_with_offspring:
			if not off_to_swap_in.name in offspring_pool.get_cluster_names():
				print('temp error in CNA_version_3.py')
				import pdb; pdb.set_trace()
				exit()
			#off_to_swap_in_index = offspring_pool.get_index(off_to_swap_in.name)
			force_replacement.append((pop_to_swap_out.name,off_to_swap_in.name))
		force_replacement.sort(key=lambda x:x[1],reverse=True) # force replace from highest to lowest index
		#for pop_to_swap_out_name, off_to_swap_in_name in force_replacement:
		#	print(off_to_swap_in_name)
		return offspring_to_remove, force_replacement
		########################################################################################################

	def structures_to_compare_generator(self,offspring_pool):
		for offspring in offspring_pool:
			offspring_name = offspring.name
			for cluster_pop in self.population:
				cluster_pop_name = cluster_pop.name
				yield (offspring_name, cluster_pop_name, self.LoD_database[offspring_name],self.LoD_database[cluster_pop_name],self.similarity_settings)
		for index1 in range(len(offspring_pool)):
			off1_name = offspring_pool[index1].name
			for index2 in range(index1+1,len(offspring_pool)):
				off2_name = offspring_pool[index2].name
				yield (off1_name, off2_name, self.LoD_database[off1_name],self.LoD_database[off2_name],self.similarity_settings)

	def get_identical_structures_generation(self,offspring_pool):
		structures_to_compare = self.structures_to_compare_generator(offspring_pool)
		###############
		start_time = time.time()
		with mp.Pool(processes=self.no_of_cpus) as pool:
			results = pool.map_async(self.check_if_two_clusters_are_identical_method, structures_to_compare)
			results.wait()
		data = results.get()
		end_time = time.time()
		print('Processing Cluster LoD comparisons Entries took '+str(end_time - start_time)+' seconds.')
		for compared_cluster_names, is_the_same_cluster in data:
			name1, name2 = compared_cluster_names
			self.LoD_comparison_database.add(name1, name2, is_the_same_cluster)

	def assess_for_violations_message(self,swap_cluster_in_pop_with_offspring,offspring_to_remove_as_too_many_pop_in_similar_common,offspring_to_remove_due_to_pop,offspring_to_remove_due_to_other_offspring):
		"""
		This method will tell the information that offspring are removed or replaced and why. 

		:param swap_cluster_in_pop_with_offspring: The clusters in the population to swap out, and the offspring to swap in its place. These will be replaced because the offspring has a higher fitness than the cluster in the population, even though they are similar. 
		:type  swap_cluster_in_pop_with_offspring: tuple of (Organisms.GA.SCM_Predation_Operator.Cluster_Block, Organisms.GA.SCM_Predation_Operator.Cluster_Block)
		:param offspring_to_remove_as_too_many_pop_in_similar_common: Remove offspring because there are too many similar clusters to it in the population already. 
		:type  offspring_to_remove_as_too_many_pop_in_similar_common: tuple of (Organisms.GA.SCM_Predation_Operator.Cluster_Block, Organisms.GA.SCM_Predation_Operator.Cluster_Block)
		:param offspring_to_remove_due_to_pop: Remove offspring because they are similar to a cluster in the population. These offspring are less fit than their similar counterparts in the population. 
		:type  offspring_to_remove_due_to_pop: tuple of (Organisms.GA.SCM_Predation_Operator.Cluster_Block, Organisms.GA.SCM_Predation_Operator.Cluster_Block)
		:param offspring_to_remove_due_to_other_offspring: Remove offpsring because they are similar to other clusters in the Offspring_Pool. This offspring is less fit than their similar counterparts in the offspring_pool. 
		:type  offspring_to_remove_due_to_other_offspring: tuple of (Organisms.GA.SCM_Predation_Operator.Cluster_Block, Organisms.GA.SCM_Predation_Operator.Cluster_Block)

		"""
		########################################################################################################
		# Mention offspring that have been removed and due to having a similar energy to which cluster in the population
		# Error with this one
		print('1. Clusters in the Population that have been replaced by Offspring, due to the cluster in the offspring being geometrically similar and having a lower fitness than a offspring.')
		if len(swap_cluster_in_pop_with_offspring) > 0:
			print('1. Clusters in the Population that have been replaced by Offspring, due to the cluster in the offspring being geometrically similar and having a lower fitness than a offspring.')
			for pop_to_swap_out, off_to_swap_in in sorted(swap_cluster_in_pop_with_offspring,key=lambda x:x[1].name,reverse=False):
				cp_removed_name = pop_to_swap_out.name; cp_removed_fitness = pop_to_swap_out.fitness; cp_removed_energy = pop_to_swap_out.energy
				off_swapped_in_name = off_to_swap_in.name; off_swapped_in_fitness = off_to_swap_in.fitness; off_swapped_in_energy = off_to_swap_in.energy
				print(' --> Population Cluster removed '+str(cp_removed_name)+' (Energy: '+str(round(cp_removed_energy,6))+' eV; fitness: '+str(round(cp_removed_fitness,6))+') {Replaced with Offspring '+str(off_swapped_in_name)+' (Energy: '+str(round(off_swapped_in_energy,6))+' eV; fitness: '+str(round(off_swapped_in_fitness,6))+')}.')
		else:
			print('1. No clusters in the population have been replaced by offspring in the offspring_pool list')
		print('*****************************')
		########################################################################################################
		# Mention offspring that have been removed and due to having a similar energy to which cluster in the population
		print('2. These offspring have a better fitness and are geometrically similar to two or more clusters in the population.')
		if len(offspring_to_remove_as_too_many_pop_in_similar_common) > 0:
			print('2. These offspring have a better fitness and are geometrically similar to two or more clusters in the population.')
			print('   Therefore it is best to remove the offspring to avoid having geometrically similar clusters in the population.')
			print('   Below are the offspring that will be removed and the clusters in the population they are geometrically similar to.')
			for off_to_swap_out, geo_sim_pops in sorted(offspring_to_remove_as_too_many_pop_in_similar_common,key=lambda x:x[0].name,reverse=False):
				off_removed_name = off_to_swap_out.name; off_removed_fitness = off_to_swap_out.fitness; off_removed_energy = off_to_swap_out.energy
				geo_sim_pop1, geo_sim_pop2 = geo_sim_pops
				geo_sim_pop1_name = geo_sim_pop1.name; geo_sim_pop1_fitness = geo_sim_pop1.fitness; geo_sim_pop1_energy = geo_sim_pop1.energy
				geo_sim_pop2_name = geo_sim_pop2.name; geo_sim_pop2_fitness = geo_sim_pop2.fitness; geo_sim_pop2_energy = geo_sim_pop2.energy
				print(' --> Offspring removed '+str(off_removed_name)+' (Energy: '+str(round(off_removed_energy,6))+' eV; fitness: '+str(round(off_removed_fitness,6))+') {Geometrically similar to (at least) clusters in population: '+str(geo_sim_pop1_name)+' (Energy: '+str(round(geo_sim_pop1_energy,6))+' eV; fitness: '+str(round(geo_sim_pop1_fitness,6))+')} and '+str(geo_sim_pop2_name)+' (Energy: '+str(round(geo_sim_pop2_energy,6))+' eV; fitness: '+str(round(geo_sim_pop2_fitness,6))+')}.')
		else:
			print('2. No offspring need to be removed from the offspring_pool list')
		print('*****************************')
		########################################################################################################
		# Mention offspring that have been removed and due to having a similar energy to which cluster in the population
		print('3. The following offspring were removed from the offspring pool (and because they are geometrically similar and have a lower fitness to which cluster in the population):')
		if len(offspring_to_remove_due_to_pop) > 0:
			print('3. The following offspring were removed from the offspring pool (and because they are geometrically similar and have a lower fitness to which cluster in the population):')
			for offspring_removed, cp_due_to in offspring_to_remove_due_to_pop:
				offspring_removed_name = offspring_removed.name; offspring_removed_fitness = offspring_removed.fitness; offspring_removed_energy = offspring_removed.energy
				cp_due_to_name = cp_due_to.name; cp_due_to_fitness = cp_due_to.fitness; cp_due_to_energy = cp_due_to.energy
				print(' --> Offspring '+str(offspring_removed_name)+' (Energy: '+str(offspring_removed_energy)+' eV; fitness: '+str(round(offspring_removed_fitness,6))+') {Similar to Population Cluster '+str(cp_due_to_name)+' (Energy: '+str(round(cp_due_to_energy,6))+' eV; fitness: '+str(round(cp_due_to_fitness,6))+')}.')
		else:
			print('3. No offspring need to be removed from the offspring_pool list')
		print('*****************************')
		########################################################################################################
		# Mention offspring that have been removed and due to having a similar energy to which other offspring in the Offspring_Pool
		print('4. Offspring that have have been removed for being geometrically similar and have a lower fitness to the rest of the other Offspring')
		if len(offspring_to_remove_due_to_other_offspring) > 0:
			print('4. Offspring that have have been removed for being geometrically similar and have a lower fitness to the rest of the other Offspring')
			for offspring_removed, due_to_offspring in offspring_to_remove_due_to_other_offspring:
				offspring_removed_name = offspring_removed.name; offspring_removed_fitness = offspring_removed.fitness; offspring_removed_energy = offspring_removed.energy
				due_to_offspring_name = due_to_offspring.name; due_to_offspring_fitness = due_to_offspring.fitness; due_to_offspring_energy = due_to_offspring.energy
				print(' --> Offspring '+str(offspring_removed_name)+' (Energy: '+str(offspring_removed_energy)+'eV; fitness: '+str(round(offspring_removed_fitness,6))+') {Similar to Offspring '+str(due_to_offspring_name)+' (Energy: '+str(due_to_offspring_energy)+' eV; fitness: '+str(round(due_to_offspring_fitness,6))+')}.')
		else:
			print('4. No offspring have been removed from the rest of the offspring_pool list for being too similar')
		print('**********************************************************')
		print('**********************************************************')
		########################################################################################################

	###########################################################################################################################################################
	########################## Methods Required for the def assess_for_violations, in the class Fitness_Factor, in Fitness_Factor.py ##########################
	###########################################################################################################################################################

	def update_database_after_natural_selection(self,offspring_not_accepted):
		for offspring_name in offspring_not_accepted:
			del self.LoD_database[offspring_name]
			self.LoD_comparison_database.remove(offspring_name)
		if not sorted(self.population.get_cluster_names()) == sorted(self.LoD_database.keys()):
			print('Error')
			import pdb; pdb.set_trace()
			exit()
		if not sorted(self.population.get_cluster_names()) == sorted(self.LoD_comparison_database.keys()):
			print('Error')
			import pdb; pdb.set_trace()
			exit()

	def reset(self):
		self.LoD_database = {}
		self.LoD_comparison_database.reset()

	########################################################################################################
	########################################################################################################
	########################################################################################################

	def add_to_database(self, collection):
		self.update_LoD_database(collection)
		self.get_identical_structures_generation(collection)

	def remove_from_database(self, cluster_names_to_remove):
		for cluster_name in cluster_names_to_remove:
			del self.LoD_database[cluster_name]
			self.LoD_comparison_database.remove(cluster_name)
		self.check_for_issues()

	########################################################################################################
	########################################################################################################
	########################################################################################################

	def check_for_issues(self):
		self.LoD_comparison_database.check_for_issues(self.population)




